# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Quantizer Utils."""

from __future__ import annotations

import os
from contextlib import contextmanager
from itertools import islice
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import datasets  # type: ignore[import]
import torch
from accelerate import cpu_offload_with_hook  # type: ignore
from tqdm import tqdm

from friendli.errors import InvalidConfigError, QuantizationError
from friendli.logging import logger
from friendli.modules.quantizer.schema.config import CalibrationDatasetConfig
from friendli.modules.quantizer.schema.data import (
    ModuleName,
    WeightActQuantResult,
    WeightOnlyQuantResult,
)


def scale_reshape(
    params: List[torch.Tensor],
) -> torch.Tensor:
    """Reshape scale/zero of quantized layers."""
    if len(params) == 1:
        t = params[0]
    else:
        t = torch.cat(params, dim=1)
    return t


def quantized_qkv_weight_reshape(
    params: List[torch.Tensor],
) -> torch.Tensor:
    """Reshape weight of quantized qkv layers."""
    assert len(params) == 3
    qkv_weight = torch.concat(
        params,
        dim=0,
    )  # [OutDim, InDim]

    return qkv_weight.to(torch.uint8)


def quantized_linear_weight_reshape(
    params: List[torch.Tensor],
) -> torch.Tensor:
    """Reshape weight of quantized linear layers."""
    assert len(params) == 1

    return params[0].to(torch.uint8)


def get_torch_data_type(data_type: str):
    """Get torch data type from Enum."""
    if data_type == "fp16":
        return torch.float16
    if data_type == "fp32":
        return torch.float32
    return torch.bfloat16


def safe_load_datasets(data_cfg: CalibrationDatasetConfig) -> datasets.Dataset:
    """Load dataset from calibration dataset config."""
    data_path = data_cfg.path_or_name
    data_split = data_cfg.split

    try:
        if os.path.exists(data_path):
            dataset = datasets.load_dataset(
                data_cfg.format,
                data_files=data_path,
                split=data_split,
            )
        else:
            data_name_parts = data_path.split(":")
            if len(data_name_parts) == 1:
                dataset = datasets.load_dataset(data_path, split=data_split)
            elif len(data_name_parts) == 2:
                data_name, subset_name = data_name_parts
                dataset = datasets.load_dataset(
                    data_name, subset_name, split=data_split
                )
            else:
                raise InvalidConfigError(
                    "Dataset name is in invalid format. "
                    "(valid format: '<dataset_name>' or '<dataset_name>:<subset_name>')"
                )
    except ValueError as err:
        raise QuantizationError(f"datasets.load_dataset failed. {str(err)}") from err

    if not isinstance(dataset, datasets.Dataset):
        raise InvalidConfigError(
            "This dataset format is not supported for the calibration."
        )

    return dataset


T = TypeVar("T")


def batched(it: Iterator[T], n: int) -> Iterator[List[T]]:
    """Batch an iterator into lists of size n."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    while True:
        batch = list(islice(it, n))
        if not batch:
            return
        yield batch


def build_percentile_statistics(
    scale_percentile: float,
    symmetric: bool = True,
) -> Tuple[Callable, Callable, Callable]:
    """Builds the hooks for getting the max input and output activations of a model."""
    logger.info(
        "Building percentile statistics hooks. scale_percentile: (%s)",
        scale_percentile,
    )

    max_input_M1: Dict[str, torch.Tensor] = {}
    max_input_M2: Dict[str, torch.Tensor] = {}
    max_input_num: Dict[str, torch.Tensor] = {}
    max_output_M1: Dict[str, torch.Tensor] = {}
    max_output_M2: Dict[str, torch.Tensor] = {}
    max_output_num: Dict[str, torch.Tensor] = {}

    def create_hook(name: ModuleName):
        def update_stats(
            max_M1: Dict[str, torch.Tensor],
            max_M2: Dict[str, torch.Tensor],
            max_num: Dict[str, int],
            new_t: torch.Tensor,
        ) -> None:
            # Chan's method for computing mean and variance incrementally
            new_t = new_t.detach().reshape(-1, new_t.size(-1))
            new_numel = new_t.size(0)
            new_t_M1 = new_t.to(torch.float64).mean(dim=0)
            if symmetric:
                # it is assumed samples are always centered on zero
                # in the symmetric quantization scheme
                new_t_M1.zero_()
            new_t_M2 = ((new_t.to(torch.float64) - new_t_M1) ** 2).sum(dim=0)
            try:
                pre_numel = max_num[name]
                max_num[name] += new_numel
                delta = new_t_M1 - max_M1[name]
                max_M1[name] += delta * (new_numel / max_num[name])
                max_M2[name] += new_t_M2 + torch.pow(delta, 2) * (
                    pre_numel * new_numel / max_num[name]
                )
            except KeyError:
                max_num[name] = new_numel
                max_M1[name] = new_t_M1
                max_M2[name] = new_t_M2

        def hook(module, in_t_tup, out_t):  # pylint: disable=unused-argument
            with torch.no_grad():
                in_t = in_t_tup[0]
                update_stats(max_input_M1, max_input_M2, max_input_num, in_t)
                update_stats(max_output_M1, max_output_M2, max_output_num, out_t)

        return hook

    def finish_input_stats():
        return {
            name: torch.distributions.Normal(
                loc=max_input_M1[name],
                scale=torch.sqrt(max_input_M2[name] / max_input_num[name]).clip(
                    min=1e-7
                ),
            ).icdf(
                torch.Tensor([(scale_percentile / 100.0) * 0.5 + 0.5]).to(
                    max_input_M1[name].device
                )
            )
            for name in list(max_input_M1.keys())
        }

    def finish_output_stats():
        return {
            name: torch.distributions.Normal(
                loc=max_output_M1[name],
                scale=torch.sqrt(max_output_M2[name] / max_output_num[name]).clip(
                    min=1e-7
                ),
            ).icdf(
                torch.Tensor([(scale_percentile / 100.0) * 0.5 + 0.5]).to(
                    max_output_M1[name].device
                )
            )
            for name in list(max_output_M1.keys())
        }

    return finish_input_stats, finish_output_stats, create_hook


def build_max_statistics() -> Tuple[Callable, Callable, Callable]:
    """Builds the hooks for getting the max input and output activations of a model."""
    logger.info("Building max statistics hooks")
    max_input_stats: Dict[str, torch.Tensor] = {}
    max_output_stats: Dict[str, torch.Tensor] = {}

    def create_hook(name: ModuleName):
        def hook(modules, in_t_tup, out_t):  # pylint: disable=unused-argument
            in_t = in_t_tup[0]
            in_t = (
                in_t.detach().abs().reshape(-1, in_t.size(-1)).max(dim=0).values
            )  # reduce-max only leaving the hidden dim (supposing the last dim is the hidden dim)
            out_t = out_t.detach().reshape(-1, out_t.size(-1))
            out_t = out_t.abs().max(dim=0).values
            try:
                max_input_stats[name] = torch.maximum(max_input_stats[name], in_t)
            except KeyError:
                max_input_stats[name] = in_t
            try:
                max_output_stats[name] = torch.maximum(max_output_stats[name], out_t)
            except KeyError:
                max_output_stats[name] = out_t

        return hook

    def finish_input_stats():
        return max_input_stats

    def finish_output_stats():
        return max_output_stats

    return finish_input_stats, finish_output_stats, create_hook


@torch.no_grad()
def collect_stats(
    model: torch.nn.Module,
    device: str,
    dataset: datasets.Dataset,
    target_classes: Tuple[Type[torch.nn.Module], ...],
    tqdm_desc: str,
    percentile: float,
) -> Tuple[Dict[ModuleName, torch.Tensor], Dict[ModuleName, torch.Tensor]]:
    """Collects the maximum values of input and output activations of a specific model.

    Args:
        model (torch.nn.Module): The model for which we want to collect the max statistics.
        dataset (Dataset): Dataset that contains input tensors.
        target_classes (Tuple[Type[torch.nn.Module], ...]): A tuple of the target classes.

    Returns:
        A tuple of two dictionaries: (max_input_stats, max_output_stats), where:
        max_input_stats: The maximum input activation values for each module of the model.
        max_output_stats: The maximum output activation values for each module of the model.

    This function uses a forward hook to capture the maximum input and output activation values
    of the specified target_classes. The max_batch_size parameter controls the size of the input
    batches that are passed through the model.

    The function returns two dictionaries containing the maximum input and output activation
    values for each module of the model, respectively. These dictionaries can be used to calculate
    scaling factors for weight quantization and activation smoothing.

    """
    # pylint: disable=too-many-locals
    max_input_stats, max_output_stats, create_hook = (
        build_percentile_statistics(percentile)
        if percentile < 100.0
        else build_max_statistics()
    )
    name_mods = [
        (name, module)
        for name, module in model.named_modules()
        if isinstance(module, target_classes)
    ]

    removables = []
    for name, module in name_mods:
        removables.append(module.register_forward_hook(create_hook(name)))
    try:
        for inp_dict in tqdm(dataset, desc=tqdm_desc):
            inputs = torch.tensor(inp_dict["input_ids"])
            model(inputs.to(device))
    finally:
        for removable in removables:
            removable.remove()
    return max_input_stats(), max_output_stats()


def build_inps_hook():
    """Builds the hooks for getting the input and output activations of a module."""
    args_dict = {}
    kwargs_dict = {}

    def create_hook(name: ModuleName):
        def hook(m, args, kwargs, y):  # pylint: disable=unused-argument
            assert name not in args_dict
            assert name not in kwargs_dict
            # assumption: all positional arguments are torch.Tensor
            args_dict[name] = [t.detach() for t in args]
            kwargs_dict[name] = {
                k: (v.detach() if isinstance(v, torch.Tensor) else v)
                for k, v in kwargs.items()
            }

        return hook

    return args_dict, kwargs_dict, create_hook


def collect_inps(
    module: torch.nn.Module,
    module_args: Tuple[Any, ...],
    module_kwargs: Dict[str, Any],
    device: str,
    target_classes: Tuple[Type[torch.nn.Module], ...],
) -> Tuple[Dict[ModuleName, Tuple[Any]], Dict[ModuleName, Dict[str, Any]]]:
    """Collects concated input and output activations of a specific module."""
    args_dict, kwargs_dict, create_hook = build_inps_hook()
    name_mods = [
        (name, m) for name, m in module.named_modules() if isinstance(m, target_classes)
    ]

    removables = []
    for name, m in name_mods:
        removables.append(m.register_forward_hook(create_hook(name), with_kwargs=True))

    module(
        *((t.to(device) if isinstance(t, torch.Tensor) else t) for t in module_args),
        **{
            k: (v.to(device) if isinstance(v, torch.Tensor) else v)
            for k, v in module_kwargs.items()
        },
    )

    for removable in removables:
        removable.remove()

    return args_dict, kwargs_dict


def get_torch_quant_dtype(q_bit: int = 8):
    """Get torch quant data type from quant bit."""
    if q_bit == 8:
        return torch.int8
    if q_bit == 4:
        return torch.int32  # In AWQ, we use int32 to represent int4
    raise ValueError(f"Invalid quant bit: {q_bit}")


@torch.no_grad()
def get_weight_act_quant_scales(
    layer_name: str,
    input_max: torch.Tensor,
    fc_weight: torch.Tensor,
    output_max: torch.Tensor,
    device: str = "cpu",
) -> WeightActQuantResult:
    """Get the quantization scales and int8 weight for a specific layer."""
    # shape of input_max: [InChannels]
    # shape of output_max: [OutChannels]
    # shape of fc_weight: [OutChannels, InChannels]
    assert input_max.ndim == 1
    assert output_max.ndim == 1

    in_channels = input_max.size(0)
    out_channels = output_max.size(0)
    assert tuple(fc_weight.size()) == (out_channels, in_channels)

    max_int = 2 ** (8 - 1) - 1
    min_int = -(2 ** (8 - 1))

    in_scale = float(max_int) / float(input_max.detach().abs().max().item())
    weight_scale = float(max_int) / float(fc_weight.detach().abs().max().item())
    out_scale = float(max_int) / float(output_max.detach().abs().max().item())
    q_weight = (
        (fc_weight.detach().float() * weight_scale)
        .round()
        .clip(min_int, max_int)
        .to(get_torch_quant_dtype(8))
        .to(device)
    )

    return WeightActQuantResult(
        layer_name,
        q_bit=8,
        zero_point=torch.tensor(0.0),
        in_scale=torch.tensor(in_scale),
        weight_scale=torch.tensor(weight_scale),
        out_scale=torch.tensor(out_scale),
        q_weight=q_weight,
        q_group_size=-1,
    )


def get_weight_only_quant_scales(
    w: torch.Tensor,
    q_bit: int,
    q_group_size: int,
    layer_name: str = "",
    device: Union[str, torch.device] = "cpu",
) -> WeightOnlyQuantResult:
    """Return the quantization scales of weight for a specific layer."""
    org_w_shape = w.shape  # [OutDim, InDim]

    w = w.reshape(-1, q_group_size)  # [OutDim x num_groups, group_size]
    max_val = w.amax(dim=1, keepdim=True)
    min_val = w.amin(dim=1, keepdim=True)

    max_int = 2**q_bit - 1
    min_int = 0

    scales = (max_val - min_val).clamp(min=1e-5) / max_int
    zeros = (-torch.round(min_val / scales)).clamp_(min_int, max_int)

    assert torch.isnan(scales).sum() == 0

    q_weight = torch.clamp(torch.round(w / scales) + zeros, min_int, max_int)
    q_weight = q_weight.reshape(org_w_shape).detach().to(device)
    scales = (
        scales.view(org_w_shape[0], -1).transpose(0, 1).detach().to(device)
    )  # [OutDim, num_groups]
    zeros = (
        zeros.view(org_w_shape[0], -1).transpose(0, 1).detach().to(device)
    )  # [OutDim, num_groups]

    assert torch.isnan(q_weight).sum() == 0
    return WeightOnlyQuantResult(
        layer_name,
        q_bit=q_bit,
        zero_point=zeros,
        q_group_size=q_group_size,
        weight_scale=scales,
        q_weight=q_weight,
    )


def send_model_to_device(
    model: torch.nn.Module,
    device: Union[str, torch.device],
    *,
    exclude: Iterable[torch.nn.Module] = (),
):
    """Send the model and its submodules onto device except for modules designated by `exclude`."""
    exclude_set = set(exclude)

    @torch.no_grad()
    def recurse(m: torch.nn.Module):
        if m in exclude_set:
            return
        for name, p in list(m.named_parameters(recurse=False)):
            m.register_parameter(name, torch.nn.Parameter(p.to(device)))
        for name, b in list(m.named_buffers(recurse=False)):
            m.register_buffer(name, b.to(device))

        for child in m.children():
            recurse(child)

    recurse(model)


class RemovableOffloaderHook(Protocol):
    """Hook protocol for cpu offloader."""

    def offload(self) -> None:
        """Offload the associated block onto CPU."""

    def remove(self) -> None:
        """Remove this hook."""


@contextmanager
def offload_module_sequence(
    blocks: Sequence[torch.nn.Module], device: Union[str, torch.device]
):
    """Offload a sequence of torch modules automatically.

    In the beginning, all blocks are supposed to reside on CPU.
    When i-th block is called, it is loaded onto `device` on the fly.
    And at the same time, it offloads (i-1)-th block back to CPU.
    """
    module_hooks: List[RemovableOffloaderHook] = []
    if blocks:
        prev_module_hook = None
        for tf_block in blocks:
            _, module_hook = cpu_offload_with_hook(
                tf_block, device, prev_module_hook=prev_module_hook
            )
            prev_module_hook = module_hook
            module_hooks.append(module_hook)
    try:
        yield
    finally:
        for hook in module_hooks:
            hook.offload()
        for hook in module_hooks:
            hook.remove()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Checkpoint Quantizer Data Schema."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional

import torch

ModuleName = str


@dataclass
class CommonQuantResult:
    """Dataclass for quantization result per layer."""

    module_name: str
    q_bit: int
    q_group_size: int
    zero_point: torch.Tensor


@dataclass
class WeightOnlyQuantResult(CommonQuantResult):
    """Dataclass for weight-only quantization result per layer."""

    weight_scale: torch.Tensor
    q_weight: torch.Tensor


@dataclass
class WeightActQuantResult(WeightOnlyQuantResult):
    """Dataclass for weight-activation quantization result per layer."""

    in_scale: torch.Tensor
    out_scale: torch.Tensor


@dataclass
class QuantInput:
    """Dataclass for int8 quantization input of each layer in transformer block."""

    weight: torch.Tensor  # [OutDim, InDim]
    name: ModuleName
    start_offset: Optional[int]  # start offset of the weight tensor along the out_dim
    end_offset: Optional[int]  # end offset of the weight tensor along the out_dim
    sort_fn: Optional[
        Callable[[torch.Tensor], torch.Tensor]
    ] = None  # sort function for max_output_stats


@dataclass
class TFQuantInputs:  # pylint: disable=too-many-instance-attributes
    """Dataclass for int8 quantization input per transformer block."""

    layer_index: int
    block: torch.nn.Module
    q: QuantInput
    k: QuantInput
    v: QuantInput
    attn_fc: QuantInput
    ff1: QuantInput
    ff2: QuantInput


@dataclass
class TFQuantResults:  # pylint: disable=too-many-instance-attributes
    """Dataclass for int8 quantization result per a transformer block."""

    layer_prefix_with_index: str
    block: torch.nn.Module
    q: CommonQuantResult
    k: CommonQuantResult
    v: CommonQuantResult
    attn_fc: CommonQuantResult
    ff1: CommonQuantResult
    ff2: CommonQuantResult



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Interface for resource management SDK."""

# pylint: disable=redefined-builtin

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Type, TypeVar, Union, overload

import pydantic

from friendli.client.graphql.base import GqlClient
from friendli.context import get_current_project_id, get_current_team_id
from friendli.errors import AuthorizationError
from friendli.utils.compat import model_parse

_Resource = TypeVar("_Resource", bound=pydantic.BaseModel)
_ResourceId = TypeVar("_ResourceId")
_Client = TypeVar("_Client", bound=GqlClient)


@dataclass
class ResourceApi(ABC, Generic[_Client, _Resource, _ResourceId]):
    """Abstract class for resource APIs."""

    client: _Client

    @property
    @abstractmethod
    def _resource_model(self) -> Type[_Resource]:
        """Model type of resource."""

    @abstractmethod
    def create(self, *args, **kwargs) -> _Resource:
        """Creates a resource."""

    @abstractmethod
    def get(self, eid: _ResourceId, *args, **kwargs) -> _Resource:
        """Gets a specific resource."""

    @abstractmethod
    def list(self, *args, **kwargs) -> List[_Resource]:
        """Lists reousrces."""

    @overload
    def _model_parse(self, data: Dict[str, Any]) -> _Resource:
        ...

    @overload
    def _model_parse(self, data: List[Dict[str, Any]]) -> List[_Resource]:
        ...

    def _model_parse(
        self, data: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> Union[_Resource, List[_Resource]]:
        if isinstance(data, list):
            return [model_parse(self._resource_model, entry["node"]) for entry in data]
        return model_parse(self._resource_model, data)

    def _get_project_id(self) -> str:
        project_id = get_current_project_id()
        if project_id is None:
            raise AuthorizationError(
                "Project to run as is not configured. Set 'FRIENDLI_PROJECT' "
                "environment variable, or use a CLI command 'friendli project switch'."
            )
        return project_id

    def _get_team_id(self) -> str:
        team_id = get_current_team_id()
        if team_id is None:
            raise AuthorizationError(
                "Team to run as is not configured. Set 'FRIENDLI_TEAM' environment "
                "variable, or use a CLI command 'friendli team switch'."
            )
        return team_id



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Organization CLI."""

from __future__ import annotations

import typer

from friendli.client.graphql.user import UserGqlClient
from friendli.context import get_current_team_id, set_current_team_id
from friendli.formatter import TableFormatter
from friendli.utils.decorator import check_api
from friendli.utils.format import secho_error_and_exit

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
)

team_table_formatter = TableFormatter(
    name="Teams",
    fields=["node.id", "node.name", "node.dedicated.plan"],
    headers=["ID", "Name", "Dedicated Plan"],
    substitute_exact_match_only=False,
)


@app.command("list")
@check_api
def list_teams():
    """List teams."""
    client = UserGqlClient()
    teams = client.get_teams()
    current_team_id = get_current_team_id()

    for team in teams:
        if current_team_id is not None and team["node"]["id"] == current_team_id:
            team["node"]["id"] = f"[bold green]* {team['node']['id']}"
            team["node"]["name"] = f"[bold green]{team['node']['name']}"
            if team["node"]["dedicated"] is not None:
                team["node"]["dedicated"][
                    "plan"
                ] = f"[bold green]{team['node']['dedicated']['plan']}"
            else:
                team["node"]["dedicated"] = {"plan": "[bold green]-"}
        else:
            team["node"]["id"] = f"  {team['node']['id']}"

    team_table_formatter.render(teams)


@app.command("switch")
def switch_team(team_id: str = typer.Argument(..., help="ID of team to switch.")):
    """Switch current team context to run as."""
    client = UserGqlClient()
    accessible_team_ids = client.get_team_ids()
    if team_id not in accessible_team_ids:
        secho_error_and_exit(f"'{team_id}' is not valid team ID.")

    set_current_team_id(team_id)
    typer.secho(
        f"Team context is switched to '{team_id}'.",
        fg=typer.colors.GREEN,
    )



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

# pylint: disable=too-many-locals, too-many-arguments

"""Generates Text Completions via CLI."""

from __future__ import annotations

from typing import List, Optional

import typer

from friendli.sdk.client import Friendli
from friendli.utils.compat import model_dump
from friendli.utils.decorator import check_api, check_api_params

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
)


@app.command()
@check_api_params
@check_api
def create(
    prompt: str = typer.Option(
        ...,
        "--prompt",
        "-p",
        help="The input text to generate completion for.",
    ),
    model: Optional[str] = typer.Option(
        None,
        "--model",
        "-m",
        help=(
            "The model to use for chat completions. "
            "See https://docs.friendli.ai/guides/serverless_endpoints/pricing for more "
            "about available models and pricing."
        ),
    ),
    endpoint_id: Optional[str] = typer.Option(
        None,
        "--endpoint-id",
        "-e",
        help="Dedicated endpoint ID to send request.",
    ),
    n: Optional[int] = typer.Option(
        None,
        "--n",
        "-n",
        min=1,
        help="The number of results to generate.",
    ),
    max_tokens: Optional[int] = typer.Option(
        None,
        "--max-tokens",
        "-M",
        min=1,
        help="The maximum number of tokens to generate.",
    ),
    stop: Optional[List[str]] = typer.Option(
        None,
        "--stop",
        "-S",
        help=(
            "When one of the stop phrases appears in the generation result, the API "
            "will stop generation. The stop phrases are excluded from the result. "
            "Repeat this option to use multiple stop phrases."
        ),
    ),
    temperature: Optional[float] = typer.Option(
        None,
        "--temperature",
        "-T",
        min=0,
        help="Sampling temperature. non-zero positive numbers are allowed.",
    ),
    top_p: Optional[float] = typer.Option(
        None,
        "--top-p",
        "-P",
        help="Tokens comprising the top top_p probability mass are kept for sampling.",
        min=0,
        max=1,
    ),
    frequency_penalty: Optional[float] = typer.Option(
        None,
        "--frequency-penalty",
        "-fp",
        min=-2,
        max=2,
        help=(
            "Positive values penalizes tokens that have been sampled, taking into "
            "account their frequency in the preceding text. This penalization "
            "diminishes the model's tendency to reproduce identical lines verbatim."
        ),
    ),
    presence_penalty: Optional[float] = typer.Option(
        None,
        "--presence-penalty",
        "-pp",
        min=-2.0,
        max=2.0,
        help=(
            "Positive values penalizes tokens that have been sampled at least once in "
            "the existing text."
        ),
    ),
    enable_stream: bool = typer.Option(
        False,
        "--stream",
        "-s",
        help="Whether to stream generation result.",
    ),
    token: Optional[str] = typer.Option(
        None, "--token", "-t", help="Personal access token for auth."
    ),
    team_id: Optional[str] = typer.Option(None, "--team", help="ID of team to run as."),
):
    """Creates text completions."""
    client = Friendli(token=token, team_id=team_id, endpoint_id=endpoint_id)
    if enable_stream:
        stream = client.completions.create(
            stream=True,
            model=model,
            prompt=prompt,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            max_tokens=max_tokens,
            n=n,
            stop=stop,
            temperature=temperature,
            top_p=top_p,
        )
        for chunk in stream:
            if n is not None and n > 1:
                typer.echo(model_dump(chunk))
            else:
                typer.echo(chunk.text, nl=False)
    else:
        completion = client.completions.create(
            stream=False,
            model=model,
            prompt=prompt,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            max_tokens=max_tokens,
            n=n,
            stop=stop,
            temperature=temperature,
            top_p=top_p,
        )
        typer.echo(model_dump(completion))



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI Formatting Utilities."""

from __future__ import annotations

import os
import re
from datetime import datetime, timedelta, timezone
from typing import NoReturn, Optional

import typer


# pylint: disable=line-too-long
def datetime_to_pretty_str(past: datetime, long_list: bool = False):
    """Beautify datetime."""
    cur = datetime.now().astimezone()
    delta = cur - past
    if long_list:
        if delta < timedelta(minutes=1):
            time_str = f"{delta.seconds % 60}s ago"
        elif delta < timedelta(hours=1):
            time_str = (
                f"{round((delta.seconds % 3600) / 60)}m {delta.seconds % 60}s ago"
            )
        elif delta < timedelta(days=1):
            time_str = f"{delta.seconds // 3600}h {round((delta.seconds % 3600) / 60)}m {delta.seconds % 60}s ago"
        elif delta < timedelta(days=3):
            time_str = (
                f"{delta.days}d {delta.seconds // 3600}h "
                f"{round((delta.seconds % 3600) / 60)}m ago"
            )
        else:
            time_str = past.astimezone(tz=cur.tzinfo).strftime("%Y-%m-%d %H:%M:%S")
    else:
        if delta < timedelta(hours=1):
            time_str = f"{round((delta.seconds % 3600) / 60)} mins ago"
        elif delta < timedelta(days=1):
            time_str = f"{round(delta.seconds / 3600)} hours ago"
        else:
            time_str = f"{delta.days + round(delta.seconds / (3600 * 24))} days ago"

    return time_str


def timedelta_to_pretty_str(delta: timedelta, long_list: bool = False) -> str:
    """Beautify timedelta."""
    if long_list:
        if delta < timedelta(minutes=1):
            time_str = f"{(delta.seconds % 60)}s"
        elif delta < timedelta(hours=1):
            time_str = f"{(delta.seconds % 3600) // 60}m {(delta.seconds % 60)}s"
        elif delta < timedelta(days=1):
            time_str = f"{delta.seconds // 3600}h {(delta.seconds % 3600) // 60}m {(delta.seconds % 60)}s"
        else:
            time_str = (
                f"{delta.days}d {delta.seconds // 3600}h "
                f"{(delta.seconds % 3600) // 60}m {delta.seconds % 60}s"
            )
    else:
        if delta < timedelta(hours=1):
            time_str = f"{round((delta.seconds % 3600) / 60)} mins"
        elif delta < timedelta(days=1):
            time_str = f"{round(delta.seconds / 3600)} hours"
        else:
            time_str = f"{delta.days + round(delta.seconds / (3600 * 24))} days"

    return time_str


def secho_error_and_exit(text: str, color: str = typer.colors.RED) -> NoReturn:
    """Print error and exit."""
    typer.secho(text, err=True, fg=color)
    raise typer.Exit(1)


def get_remaining_terminal_columns(occupied: int) -> int:
    """Calculate the remaining terminal column count."""
    return os.get_terminal_size().columns - occupied


def utc_to_local(dt: datetime) -> datetime:
    """Convert datatime in the UTC timezone to the local timezone."""
    return dt.replace(tzinfo=timezone.utc).astimezone(tz=None)


def datetime_to_simple_string(dt: datetime) -> str:
    """Convert a datatime to simple string."""
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def _regex_parse(pattern: str, s: str) -> Optional[str]:
    match = re.search(pattern, s)
    if match:
        return match.group()
    return None


def extract_datetime_part(s: str) -> Optional[str]:
    """Extracts the datetime portion in the format "YYYY-MM-DD--HH" from the input string `s`.

    Args:
        s: A string containing a datetime in the format "YYYY-MM-DD--HH".

    Returns:
        str: A string representing the datetime in the format "YYYY-MM-DD--HH", if found in the input string `s`. If the datetime is not found, None is returned.

    """
    pattern = r"\d{4}-\d{2}-\d{2}--\d{2}"
    return _regex_parse(pattern, s)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""CLI App Settings."""


class Settings:
    """CLI app settings."""

    access_token_cookie_key = ""
    refresh_token_cookie_key = ""


class ProductionSettings:
    """Production CLI app settings."""

    access_token_cookie_key = "sAccessTokenProduction"
    refresh_token_cookie_key = "sRefreshTokenProduction"


class StagingSettings:
    """Staging CLI app settings."""

    access_token_cookie_key = "sAccessTokenStaging"
    refresh_token_cookie_key = "sRefreshTokenStaging"


class DevSettings:
    """Dev CLI app settings."""

    access_token_cookie_key = "sAccessTokenDev"
    refresh_token_cookie_key = "sRefreshTokenDev"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli LlamaForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterator, List, Tuple, Type, cast

import torch

from friendli.modules.converter.base import DECODER_PREFIX
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.quantizer.awq.base import AWQHook
from friendli.modules.quantizer.schema.config import AWQConfig
from friendli.modules.quantizer.schema.data import (
    ModuleName,
    QuantInput,
    TFQuantInputs,
    TFQuantResults,
    WeightOnlyQuantResult,
)
from friendli.modules.quantizer.utils import (
    get_weight_only_quant_scales,
    quantized_linear_weight_reshape,
    scale_reshape,
)


@dataclass
class LlamaTFQuantInputs(TFQuantInputs):
    """Dataclass for quantization input per layer in LlamaForCausalLM."""

    ff_gate: QuantInput


@dataclass
class LlamaTFQuantResults(TFQuantResults):
    """Dataclass for quantization result per layer in LlamaForCausalLM."""

    ff_gate: WeightOnlyQuantResult


class AWQLlamaHook(AWQHook):
    """AWQ Hook for LlamaForCausalLM."""

    def __init__(self, quant_config, converter):
        """Initialize AWQLlamaHook."""
        super().__init__(quant_config, converter)
        config = converter.config
        self.data_type = converter.data_type
        self.num_attention_heads = config.num_attention_heads
        if config.num_key_value_heads is None:
            self.num_kv_attention_heads = self.num_attention_heads
        else:
            self.num_kv_attention_heads = config.num_key_value_heads
        self.hidden_size = config.hidden_size
        self.head_size = self.hidden_size // self.num_attention_heads
        self.rotary_dim = self.head_size
        self.scale_attn_fc = self.num_attention_heads == self.num_kv_attention_heads

    def add_pre_scaler(self, model: torch.nn.Module) -> torch.nn.Module:
        """Adds scaler to LlamaForCausalLM."""
        return model

    def get_inspect_module_types(
        self, block: torch.nn.Module
    ) -> Tuple[type[torch.nn.Module], ...]:
        """Returns the layer types in inspected blocks."""
        return (type(block.self_attn), type(block.mlp))

    def iter_inspect_modules(
        self,
        block: torch.nn.Module,
    ) -> Iterator[
        Tuple[
            List[torch.nn.Module],
            List[Tuple[ModuleName, torch.nn.Linear]],
            torch.nn.Module,
            ModuleName,
        ]
    ]:
        """Returns iterator of layers in blocks."""
        # qkv proj
        yield (
            [block.input_layernorm],
            [
                ("self_attn.q_proj", block.self_attn.q_proj),
                ("self_attn.k_proj", block.self_attn.k_proj),
                ("self_attn.v_proj", block.self_attn.v_proj),
            ],
            block.self_attn,
            "self_attn",
        )
        # attn out proj
        if self.scale_attn_fc:
            yield (
                [block.self_attn.v_proj],
                [("self_attn.o_proj", block.self_attn.o_proj)],
                block.self_attn.o_proj,
                "self_attn.o_proj",
            )
        # ff1
        yield (
            [block.post_attention_layernorm],
            [
                ("mlp.up_proj", block.mlp.up_proj),
                ("mlp.gate_proj", block.mlp.gate_proj),
            ],
            block.mlp,
            "mlp",
        )
        # ff2
        yield (
            [block.mlp.up_proj],
            [("mlp.down_proj", block.mlp.down_proj)],
            block.mlp.down_proj,
            "mlp.down_proj",
        )

    def iter_tf_quant_inputs(self, model: torch.nn.Module) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of LlamaForCausalLM."""
        for index, decoder_layer in enumerate(
            self.get_tf_blocks(model)  # type: ignore[union-attr, arg-type]
        ):
            self_attn = decoder_layer.self_attn
            q_weight, k_weight, v_weight = (
                self.converter.qkv_weight_reshape(
                    [
                        self_attn.q_proj.weight,
                        self_attn.k_proj.weight,
                        self_attn.v_proj.weight,
                    ]
                )
                .transpose(0, 1)
                .split(
                    [
                        self.converter.decoder_num_attention_heads
                        * self.converter.decoder_head_size,
                        self.converter.decoder_num_kv_attention_heads
                        * self.converter.decoder_head_size,
                        self.converter.decoder_num_kv_attention_heads
                        * self.converter.decoder_head_size,
                    ],
                    dim=0,
                )
            )
            fc1 = decoder_layer.mlp.up_proj
            ff_gate = decoder_layer.mlp.gate_proj
            fc2 = decoder_layer.mlp.down_proj

            yield LlamaTFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    q_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.q_proj",
                    None,
                    None,
                ),
                k=QuantInput(
                    k_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.k_proj",
                    None,
                    None,
                ),
                v=QuantInput(
                    v_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.v_proj",
                    None,
                    None,
                ),
                attn_fc=QuantInput(
                    self_attn.o_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.o_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.up_proj",
                    None,
                    None,
                ),
                ff_gate=QuantInput(
                    ff_gate.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.gate_proj",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.down_proj",
                    None,
                    None,
                ),
            )

    def get_quant_result(
        self,
        quant_input: TFQuantInputs,
        **kwargs: Any,
    ) -> TFQuantResults:
        """Get quantization result for a specific layer in LlamaForCausalLM."""
        awq_config = cast(AWQConfig, self.quant_config)

        def get_scale(quant_input: QuantInput) -> WeightOnlyQuantResult:
            weight, name, start, end = (
                quant_input.weight,
                quant_input.name,
                quant_input.start_offset,
                quant_input.end_offset,
            )
            weight = weight.to(awq_config.device)

            return get_weight_only_quant_scales(
                layer_name=name,
                w=weight[start:end],
                q_bit=awq_config.awq_args.quant_bit,
                q_group_size=awq_config.awq_args.quant_group_size,
            )

        quant_input = cast(LlamaTFQuantInputs, quant_input)
        return LlamaTFQuantResults(
            layer_prefix_with_index=f"{self.quantized_layer_prefix}{quant_input.layer_index}.",
            block=quant_input.block,
            q=get_scale(quant_input.q),
            k=get_scale(quant_input.k),
            v=get_scale(quant_input.v),
            attn_fc=get_scale(quant_input.attn_fc),
            ff1=get_scale(quant_input.ff1),
            ff_gate=get_scale(quant_input.ff_gate),
            ff2=get_scale(quant_input.ff2),
        )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in LlamaForCausalLM."""
        return (torch.nn.Linear,)

    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the transformer blocks in LlamaForCausalLM."""
        return model.model.layers

    @property
    def quantized_param_names(self) -> List[str]:
        """Returns the parameter names in LlamaForCausalLM."""
        param_names = super().quantized_param_names
        for i in range(self.converter.decoder_layer_num):
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            param_names.append(
                f"{converted_prefix}mlp/c_gate/weight:0",
            )
        return param_names

    @property
    def modified_layers_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for modified layers."""
        return []

    @property
    def avoid_clipping_layer_names(self) -> List[str]:
        """Returns the layer names which should be avoided for AWQ clipping."""
        return ["q_proj", "k_proj"]

    @property
    def quantized_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the convert_info_list for quantized layers."""
        convert_info_list = super().quantized_convert_info_list
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.quantized_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"

            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff_gate.weight_scale"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}mlp/c_gate/awq/scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff_gate.zeros"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}mlp/c_gate/awq/zero:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff_gate.weight"],
                        data_type=self.quant_dtype,
                        converted_name=f"{converted_prefix}mlp/c_gate/awq/weight:0",
                        reshape_fn=quantized_linear_weight_reshape,
                    ),
                ]
            )
        return convert_info_list



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

from __future__ import annotations

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

DESCRIPTOR: _descriptor.FileDescriptor

class V1CompletionsRequest(_message.Message):
    __slots__ = [
        "stream",
        "prompt",
        "tokens",
        "timeout_microseconds",
        "max_tokens",
        "max_total_tokens",
        "min_tokens",
        "min_total_tokens",
        "n",
        "num_beams",
        "length_penalty",
        "early_stopping",
        "no_repeat_ngram",
        "encoder_no_repeat_ngram",
        "repetition_penalty",
        "encoder_repetition_penalty",
        "frequency_penalty",
        "presence_penalty",
        "temperature",
        "top_k",
        "top_p",
        "stop",
        "stop_tokens",
        "seed",
        "token_index_to_replace",
        "embedding_to_replace",
        "beam_search_type",
        "beam_compat_pre_normalization",
        "beam_compat_no_post_normalization",
        "bad_words",
        "bad_word_tokens",
        "include_output_logits",
        "include_output_logprobs",
        "forced_output_tokens",
        "eos_token",
    ]

    class BeamSearchType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        DETERMINISTIC: _ClassVar[V1CompletionsRequest.BeamSearchType]
        STOCHASTIC: _ClassVar[V1CompletionsRequest.BeamSearchType]
        NAIVE_SAMPLING: _ClassVar[V1CompletionsRequest.BeamSearchType]
    DETERMINISTIC: V1CompletionsRequest.BeamSearchType
    STOCHASTIC: V1CompletionsRequest.BeamSearchType
    NAIVE_SAMPLING: V1CompletionsRequest.BeamSearchType

    class TokenSequence(_message.Message):
        __slots__ = ["tokens"]
        TOKENS_FIELD_NUMBER: _ClassVar[int]
        tokens: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, tokens: _Optional[_Iterable[int]] = ...) -> None: ...
    STREAM_FIELD_NUMBER: _ClassVar[int]
    PROMPT_FIELD_NUMBER: _ClassVar[int]
    TOKENS_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_MICROSECONDS_FIELD_NUMBER: _ClassVar[int]
    MAX_TOKENS_FIELD_NUMBER: _ClassVar[int]
    MAX_TOTAL_TOKENS_FIELD_NUMBER: _ClassVar[int]
    MIN_TOKENS_FIELD_NUMBER: _ClassVar[int]
    MIN_TOTAL_TOKENS_FIELD_NUMBER: _ClassVar[int]
    N_FIELD_NUMBER: _ClassVar[int]
    NUM_BEAMS_FIELD_NUMBER: _ClassVar[int]
    LENGTH_PENALTY_FIELD_NUMBER: _ClassVar[int]
    EARLY_STOPPING_FIELD_NUMBER: _ClassVar[int]
    NO_REPEAT_NGRAM_FIELD_NUMBER: _ClassVar[int]
    ENCODER_NO_REPEAT_NGRAM_FIELD_NUMBER: _ClassVar[int]
    REPETITION_PENALTY_FIELD_NUMBER: _ClassVar[int]
    ENCODER_REPETITION_PENALTY_FIELD_NUMBER: _ClassVar[int]
    FREQUENCY_PENALTY_FIELD_NUMBER: _ClassVar[int]
    PRESENCE_PENALTY_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
    TOP_K_FIELD_NUMBER: _ClassVar[int]
    TOP_P_FIELD_NUMBER: _ClassVar[int]
    STOP_FIELD_NUMBER: _ClassVar[int]
    STOP_TOKENS_FIELD_NUMBER: _ClassVar[int]
    SEED_FIELD_NUMBER: _ClassVar[int]
    TOKEN_INDEX_TO_REPLACE_FIELD_NUMBER: _ClassVar[int]
    EMBEDDING_TO_REPLACE_FIELD_NUMBER: _ClassVar[int]
    BEAM_SEARCH_TYPE_FIELD_NUMBER: _ClassVar[int]
    BEAM_COMPAT_PRE_NORMALIZATION_FIELD_NUMBER: _ClassVar[int]
    BEAM_COMPAT_NO_POST_NORMALIZATION_FIELD_NUMBER: _ClassVar[int]
    BAD_WORDS_FIELD_NUMBER: _ClassVar[int]
    BAD_WORD_TOKENS_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_OUTPUT_LOGITS_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_OUTPUT_LOGPROBS_FIELD_NUMBER: _ClassVar[int]
    FORCED_OUTPUT_TOKENS_FIELD_NUMBER: _ClassVar[int]
    EOS_TOKEN_FIELD_NUMBER: _ClassVar[int]
    stream: bool
    prompt: str
    tokens: _containers.RepeatedScalarFieldContainer[int]
    timeout_microseconds: int
    max_tokens: int
    max_total_tokens: int
    min_tokens: int
    min_total_tokens: int
    n: int
    num_beams: int
    length_penalty: float
    early_stopping: bool
    no_repeat_ngram: int
    encoder_no_repeat_ngram: int
    repetition_penalty: float
    encoder_repetition_penalty: float
    frequency_penalty: float
    presence_penalty: float
    temperature: float
    top_k: int
    top_p: float
    stop: _containers.RepeatedScalarFieldContainer[str]
    stop_tokens: _containers.RepeatedCompositeFieldContainer[
        V1CompletionsRequest.TokenSequence
    ]
    seed: _containers.RepeatedScalarFieldContainer[int]
    token_index_to_replace: _containers.RepeatedScalarFieldContainer[int]
    embedding_to_replace: _containers.RepeatedScalarFieldContainer[float]
    beam_search_type: V1CompletionsRequest.BeamSearchType
    beam_compat_pre_normalization: bool
    beam_compat_no_post_normalization: bool
    bad_words: _containers.RepeatedScalarFieldContainer[str]
    bad_word_tokens: _containers.RepeatedCompositeFieldContainer[
        V1CompletionsRequest.TokenSequence
    ]
    include_output_logits: bool
    include_output_logprobs: bool
    forced_output_tokens: _containers.RepeatedScalarFieldContainer[int]
    eos_token: _containers.RepeatedScalarFieldContainer[int]
    def __init__(
        self,
        stream: bool = ...,
        prompt: _Optional[str] = ...,
        tokens: _Optional[_Iterable[int]] = ...,
        timeout_microseconds: _Optional[int] = ...,
        max_tokens: _Optional[int] = ...,
        max_total_tokens: _Optional[int] = ...,
        min_tokens: _Optional[int] = ...,
        min_total_tokens: _Optional[int] = ...,
        n: _Optional[int] = ...,
        num_beams: _Optional[int] = ...,
        length_penalty: _Optional[float] = ...,
        early_stopping: bool = ...,
        no_repeat_ngram: _Optional[int] = ...,
        encoder_no_repeat_ngram: _Optional[int] = ...,
        repetition_penalty: _Optional[float] = ...,
        encoder_repetition_penalty: _Optional[float] = ...,
        frequency_penalty: _Optional[float] = ...,
        presence_penalty: _Optional[float] = ...,
        temperature: _Optional[float] = ...,
        top_k: _Optional[int] = ...,
        top_p: _Optional[float] = ...,
        stop: _Optional[_Iterable[str]] = ...,
        stop_tokens: _Optional[
            _Iterable[_Union[V1CompletionsRequest.TokenSequence, _Mapping]]
        ] = ...,
        seed: _Optional[_Iterable[int]] = ...,
        token_index_to_replace: _Optional[_Iterable[int]] = ...,
        embedding_to_replace: _Optional[_Iterable[float]] = ...,
        beam_search_type: _Optional[
            _Union[V1CompletionsRequest.BeamSearchType, str]
        ] = ...,
        beam_compat_pre_normalization: bool = ...,
        beam_compat_no_post_normalization: bool = ...,
        bad_words: _Optional[_Iterable[str]] = ...,
        bad_word_tokens: _Optional[
            _Iterable[_Union[V1CompletionsRequest.TokenSequence, _Mapping]]
        ] = ...,
        include_output_logits: bool = ...,
        include_output_logprobs: bool = ...,
        forced_output_tokens: _Optional[_Iterable[int]] = ...,
        eos_token: _Optional[_Iterable[int]] = ...,
    ) -> None: ...



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

from __future__ import annotations

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers

DESCRIPTOR: _descriptor.FileDescriptor

class V1ChatCompletionsRequest(_message.Message):
    __slots__ = [
        "messages",
        "frequency_penalty",
        "max_tokens",
        "n",
        "presence_penalty",
        "stop",
        "stream",
        "temperature",
        "top_p",
        "timeout_microseconds",
    ]

    class Message(_message.Message):
        __slots__ = ["content", "role"]
        CONTENT_FIELD_NUMBER: _ClassVar[int]
        ROLE_FIELD_NUMBER: _ClassVar[int]
        content: str
        role: str
        def __init__(
            self, content: _Optional[str] = ..., role: _Optional[str] = ...
        ) -> None: ...
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    FREQUENCY_PENALTY_FIELD_NUMBER: _ClassVar[int]
    MAX_TOKENS_FIELD_NUMBER: _ClassVar[int]
    N_FIELD_NUMBER: _ClassVar[int]
    PRESENCE_PENALTY_FIELD_NUMBER: _ClassVar[int]
    STOP_FIELD_NUMBER: _ClassVar[int]
    STREAM_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
    TOP_P_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_MICROSECONDS_FIELD_NUMBER: _ClassVar[int]
    messages: _containers.RepeatedCompositeFieldContainer[
        V1ChatCompletionsRequest.Message
    ]
    frequency_penalty: float
    max_tokens: int
    n: int
    presence_penalty: float
    stop: _containers.RepeatedScalarFieldContainer[str]
    stream: bool
    temperature: float
    top_p: float
    timeout_microseconds: int
    def __init__(
        self,
        messages: _Optional[
            _Iterable[_Union[V1ChatCompletionsRequest.Message, _Mapping]]
        ] = ...,
        frequency_penalty: _Optional[float] = ...,
        max_tokens: _Optional[int] = ...,
        n: _Optional[int] = ...,
        presence_penalty: _Optional[float] = ...,
        stop: _Optional[_Iterable[str]] = ...,
        stream: bool = ...,
        temperature: _Optional[float] = ...,
        top_p: _Optional[float] = ...,
        timeout_microseconds: _Optional[int] = ...,
    ) -> None: ...



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli AWQ Quantizer Base."""

from __future__ import annotations

import gc
from abc import abstractmethod
from contextlib import contextmanager
from dataclasses import fields
from typing import Any, Dict, Iterator, List, Tuple, Type, cast

import datasets  # type: ignore[import]
import torch
from datasets.utils.logging import disable_progress_bar  # type: ignore[import]
from tqdm import tqdm

from friendli.enums import ModelDataType
from friendli.errors import QuantizationError
from friendli.logging import logger
from friendli.modules.converter.base import DECODER_PREFIX
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.converter.utils import get_tokenizer
from friendli.modules.quantizer.awq.utils import (
    apply_module_clip,
    apply_module_scale,
    search_module_clip,
    search_module_scale,
)
from friendli.modules.quantizer.base import AbstractQuantHook, CommonQuantizer
from friendli.modules.quantizer.layers import WeightOnlyQuantizedLinearLayer
from friendli.modules.quantizer.schema.config import AWQConfig
from friendli.modules.quantizer.schema.data import (
    ModuleName,
    QuantInput,
    TFQuantInputs,
    TFQuantResults,
    WeightOnlyQuantResult,
)
from friendli.modules.quantizer.utils import (
    collect_inps,
    get_weight_only_quant_scales,
    offload_module_sequence,
    quantized_linear_weight_reshape,
    quantized_qkv_weight_reshape,
    safe_load_datasets,
    scale_reshape,
    send_model_to_device,
)


class AWQScaler(torch.nn.Module):
    """Store AWQ scale before linear layers.

    If the linear layer is quantized, but the previous layer can't be scaled,
    then we need to store the AWQ scale in a separate module. This module
    is used to store the AWQ scale.
    """

    def __init__(self, in_dim: int):
        """Initialize AWQScaler."""
        super().__init__()
        self.scale = torch.nn.Parameter(torch.ones(in_dim))

    def forward(self, x):
        """Scale input by AWQ scale."""
        return (x / self.scale.view(1, 1, -1)).to(x.dtype)


class AWQHook(AbstractQuantHook):
    """Quantization Hook for AWQ."""

    @abstractmethod
    def iter_inspect_modules(
        self,
        block: torch.nn.Module,
    ) -> Iterator[
        Tuple[
            List[torch.nn.Module],
            List[Tuple[ModuleName, torch.nn.Linear]],
            torch.nn.Module,
            ModuleName,
        ]
    ]:
        """Returns iterator of modules to inspect for AWQ scale."""

    @abstractmethod
    def add_pre_scaler(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Add scaler for storing AWQ scale in modules."""

    @abstractmethod
    def get_inspect_module_types(
        self, block: torch.nn.Module
    ) -> Tuple[Type[torch.nn.Module], ...]:
        """Returns the type of inspect modules in transformer block."""

    def _register_pre_scaler(
        self,
        linear: torch.nn.Module,
    ) -> AWQScaler:
        """Register pre-scaler for storing AWQ scale in modules."""
        scaler = AWQScaler(linear.in_features)  # type: ignore

        def pre_scaler_hook(_, x: Tuple[Any, ...]) -> Tuple[torch.Tensor, ...]:
            return (scaler(x[0]),)

        linear.register_forward_pre_hook(pre_scaler_hook)
        return scaler

    def get_quant_result(
        self,
        quant_inputs: TFQuantInputs,
        **kwargs: Any,
    ) -> TFQuantResults:
        """Get quantization result for AWQ."""
        awq_config = cast(AWQConfig, self.quant_config)

        def get_scale(
            quant_input: QuantInput,
        ) -> WeightOnlyQuantResult:
            weight, name, start, end = (
                quant_input.weight,
                quant_input.name,
                quant_input.start_offset,
                quant_input.end_offset,
            )
            weight = weight.to(awq_config.device)

            return get_weight_only_quant_scales(
                layer_name=name,
                w=weight[start:end],
                q_bit=awq_config.awq_args.quant_bit,
                q_group_size=awq_config.awq_args.quant_group_size,
            )

        return TFQuantResults(
            layer_prefix_with_index=f"{self.quantized_layer_prefix}{quant_inputs.layer_index}.",
            block=quant_inputs.block,
            q=get_scale(quant_inputs.q),
            k=get_scale(quant_inputs.k),
            v=get_scale(quant_inputs.v),
            attn_fc=get_scale(quant_inputs.attn_fc),
            ff1=get_scale(quant_inputs.ff1),
            ff2=get_scale(quant_inputs.ff2),
        )

    @property
    def quant_dtype(self) -> ModelDataType:
        """Return the quantization dtype."""
        quant_config = cast(AWQConfig, self.quant_config)
        awq_args = quant_config.awq_args
        if awq_args.quant_bit == 4:
            return ModelDataType.INT4
        return ModelDataType.INT8

    @property
    @abstractmethod
    def avoid_clipping_layer_names(self) -> List[str]:
        """Return the layer names to avoid clipping."""

    @property
    @abstractmethod
    def modified_layers_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for modified layers."""

    @property
    def quantized_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Return the list of conversion informations for quantized layers."""
        convert_info_list = []
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.quantized_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}q.weight_scale",
                            f"{layer_prefix}k.weight_scale",
                            f"{layer_prefix}v.weight_scale",
                        ],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/awq/scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}q.zeros",
                            f"{layer_prefix}k.zeros",
                            f"{layer_prefix}v.zeros",
                        ],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/awq/zero:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}q.weight",
                            f"{layer_prefix}k.weight",
                            f"{layer_prefix}v.weight",
                        ],
                        data_type=self.quant_dtype,
                        converted_name=f"{converted_prefix}attn/c_attn/awq/weight:0",
                        reshape_fn=quantized_qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc.weight_scale"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/awq/scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc.zeros"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/awq/zero:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn_fc.weight"],
                        data_type=self.quant_dtype,
                        converted_name=f"{converted_prefix}attn/c_proj/awq/weight:0",
                        reshape_fn=quantized_linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff1.weight_scale"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/awq/scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff1.zeros"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/awq/zero:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff1.weight"],
                        data_type=self.quant_dtype,
                        converted_name=f"{converted_prefix}mlp/c_fc/awq/weight:0",
                        reshape_fn=quantized_linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2.weight_scale"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/awq/scale:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2.zeros"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/awq/zero:0",
                        reshape_fn=scale_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ff2.weight"],
                        data_type=self.quant_dtype,
                        converted_name=f"{converted_prefix}mlp/c_proj/awq/weight:0",
                        reshape_fn=quantized_linear_weight_reshape,
                    ),
                ]
            )
        return convert_info_list


class AWQQuantizer(CommonQuantizer):
    """Quantizer for AWQ."""

    def check_config(self) -> None:
        """Check if the AWQ quantization config is valid."""
        super().check_config()
        quant_config = cast(AWQConfig, self.quant_config)
        awq_args = quant_config.awq_args
        if awq_args.quant_bit not in [4, 8]:
            raise QuantizationError(
                f"Invalid quant_bit {awq_args.quant_bit} for AWQ."
                "You can only use 4 or 8 bit for AWQ."
            )
        if awq_args.quant_group_size not in [64]:
            raise QuantizationError(
                f"Invalid quant_group_size {awq_args.quant_group_size} for AWQ."
                "You can only use 64 for AWQ."
            )

    def get_calib_dataset(self) -> datasets.Dataset:
        """Get calibration dataset for AWQ."""
        data_cfg = self.quant_config.calibration_dataset
        tokenizer = get_tokenizer(self.converter.config.name_or_path)
        dataset = safe_load_datasets(data_cfg)

        def preprocess(sample) -> Dict[str, Any]:
            """Preprocess dataset for AWQ."""
            return {"input_ids": tokenizer(sample).input_ids}

        disable_progress_bar()
        dataset = (
            dataset.shuffle(self.quant_config.seed)
            .filter(
                lambda sample: len(tokenizer(sample)) <= data_cfg.max_length,
                input_columns=data_cfg.lookup_column_name,
            )
            .map(function=preprocess, input_columns=data_cfg.lookup_column_name)
            .filter(
                lambda sample: torch.tensor(sample).numel() != 0,
                input_columns="input_ids",
            )
            .select(range(data_cfg.num_samples))
        )

        return dataset

    def get_batched_samples(self):
        """Get batched samples from dataset."""
        dataset = self.get_calib_dataset()

        samples = []
        for sample in dataset["input_ids"]:
            samples.append(torch.tensor(sample))

        seqlen = self.quant_config.calibration_dataset.max_length
        batched_samples = torch.cat(samples)
        if len(batched_samples) // seqlen == 0:
            return batched_samples.unsqueeze(0)

        batched_samples = [
            batched_samples[i * seqlen : (i + 1) * seqlen].unsqueeze(0)
            for i in range(len(batched_samples) // seqlen)
        ]
        batched_samples = torch.cat(batched_samples, dim=0)

        return batched_samples

    def _apply_awq_scale_clip_block(
        self,
        block: torch.nn.Module,
        block_args: Tuple[Any, ...],
        block_kwargs: Dict[str, Any],
    ) -> None:
        """Search AWQ scale, clipping range and Apply them into a transformer block."""
        # pylint: disable=too-many-locals

        inpsected_mod_types = cast(AWQHook, self.hook).get_inspect_module_types(block)
        args_dict, kwargs_dict = collect_inps(
            block,
            block_args,
            block_kwargs,
            self.quant_config.device,
            tuple([*self.hook.get_linear_layer_types(), *inpsected_mod_types]),
        )
        awq_args = cast(AWQConfig, self.quant_config).awq_args
        for prev_ops, linear_tuples, module2inspect, module2inspect_name in cast(
            AWQHook, self.hook
        ).iter_inspect_modules(block):
            linear_inp = args_dict[linear_tuples[0][0]][0]
            linear_layers = [linear for _, linear in linear_tuples]

            scales = search_module_scale(
                module2inspect,
                args_dict[module2inspect_name],
                kwargs_dict[module2inspect_name],
                linear_layers,
                linear_inp,
                awq_args.quant_group_size,
                awq_args.quant_bit,
            )

            apply_module_scale(
                prev_ops,
                linear_layers,
                scales.to(self.quant_config.device),
            )

            for name, _ in linear_tuples:
                assert len(args_dict[name]) == 1
                assert torch.equal(args_dict[name][0], linear_inp)
                args_dict[name] = (args_dict[name][0].div(scales.view(1, -1)),)

        named_linears = {
            name: m
            for name, m in block.named_modules()
            if isinstance(m, torch.nn.Linear)
        }
        for name, linear in named_linears.items():
            if any(
                (
                    avoid in name
                    for avoid in cast(AWQHook, self.hook).avoid_clipping_layer_names
                )
            ):
                continue
            max_val = search_module_clip(
                linear.weight,
                args_dict[name][0],
                awq_args.quant_group_size,
                awq_args.quant_bit,
                n_sample_token=self.quant_config.calibration_dataset.num_samples,
            )
            apply_module_clip(
                max_val.to(self.quant_config.device),
                linear,
            )

    def get_input_kwargs_tf_blocks(
        self,
        model: torch.nn.Module,
    ) -> Tuple[List[Tuple[Any, ...]], List[Dict[str, Any]]]:
        """Gather input tensor and kwargs from the designated pytorch module."""
        block_args = []
        block_kwargs = []

        num_tf_blocks = len(self.hook.get_tf_blocks(model))
        progress_bar = tqdm(
            range(num_tf_blocks),
            total=num_tf_blocks,
            desc="Collect args for transformer blocks..",
        )

        def hook(m, args, kwargs):  # pylint: disable=unused-argument
            block_args.append(
                tuple(
                    (t.detach().cpu() if isinstance(t, torch.Tensor) else t)
                    for t in args
                )
            )
            block_kwargs.append(
                {
                    k: (v.detach().cpu() if isinstance(v, torch.Tensor) else v)
                    for k, v in kwargs.items()
                }
            )
            progress_bar.update()

        removables = []
        for tf_block in self.hook.get_tf_blocks(model):
            removables.append(
                tf_block.register_forward_pre_hook(hook, with_kwargs=True)
            )

        batched_samples = self.get_batched_samples()
        model(batched_samples.to(self.quant_config.device), use_cache=False)

        for removable in removables:
            removable.remove()

        return block_args, block_kwargs

    def get_attributes(self) -> Dict[str, Any]:
        """Return the attributes of the converted model."""
        attributes = self.converter.get_attributes()
        awq_args = cast(AWQConfig, self.quant_config).awq_args
        attributes["quant_scheme"] = self.quant_config.mode.value  # awq
        attributes["quant_group_size"] = awq_args.quant_group_size
        attributes["quant_bit"] = awq_args.quant_bit
        return attributes

    @contextmanager
    def _try_offload_model(self, model: torch.nn.Module):
        if not self.quant_config.offload:
            logger.info("AWQ offloading not enabled. Skipping.")
            model.to(self.quant_config.device)
            yield
        else:
            logger.info("AWQ offloading enabled.")
            tf_blocks = self.hook.get_tf_blocks(model)
            send_model_to_device(model, self.quant_config.device, exclude=tf_blocks)
            with offload_module_sequence(tf_blocks, self.quant_config.device):
                yield

    @torch.no_grad()
    def _apply_awq_scale_clip(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Search AWQ scale, clipping range and Apply them into model."""
        # pylint: disable=too-many-locals
        model.eval()
        with self._try_offload_model(model):
            tf_blocks = self.hook.get_tf_blocks(model)
            block_args, block_kwargs = self.get_input_kwargs_tf_blocks(model)

            gc.collect()
            torch.cuda.empty_cache()

            for block, args, kwargs in tqdm(
                zip(
                    tf_blocks,
                    block_args,
                    block_kwargs,
                ),
                total=len(tf_blocks),
                desc="Search and Apply AWQ Scale, Clip range..",
            ):
                self._apply_awq_scale_clip_block(block, args, kwargs)
                gc.collect()
                torch.cuda.empty_cache()

    @torch.no_grad()
    def pre_quantize(
        self,
        model: torch.nn.Module,
    ) -> None:
        """Pre-procedure that should be called before quantize() is called."""
        model = cast(AWQHook, self.hook).add_pre_scaler(model)
        self._apply_awq_scale_clip(model)

    @torch.no_grad()
    def quantize(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Quantize model with AWQ."""
        model.eval()
        for quant_input in tqdm(
            self.hook.iter_tf_quant_inputs(model),
            total=len(self.hook.get_tf_blocks(model)),
            desc="Quantize model..",
        ):
            quant_result = cast(AWQHook, self.hook).get_quant_result(
                quant_input, quant_config=cast(AWQConfig, self.quant_config)
            )
            for field in fields(quant_result):
                layer_quant_result = getattr(quant_result, field.name)
                if isinstance(layer_quant_result, WeightOnlyQuantResult):
                    layer = model.get_submodule(layer_quant_result.module_name)
                    q_layer = WeightOnlyQuantizedLinearLayer.from_layer(
                        layer, layer_quant_result
                    )
                    quant_result.block.add_module(field.name, q_layer)

        return model



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli T5 Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Dict, List, cast

import torch
from transformers import T5Config  # type: ignore[import]

from friendli.enums import ModelDataType
from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import (
    DECODER_PREFIX,
    ENCODER_PREFIX,
    EncoderDecoderConverter,
)
from friendli.modules.converter.schema import ConvertInfo


class T5Converter(EncoderDecoderConverter):
    """T5ForConditionalGeneration Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if T5 architectures' config can be converted to Friendli format."""
        super().check_config()
        try:
            if not (
                cast(T5Config, self.config).is_gated_act
                ^ cast(T5Config, self.config).tie_word_embeddings
            ):
                raise NotSupportedCheckpointError(
                    invalid_option=f"'is_gated_act={cast(T5Config, self.config).is_gated_act}'and "
                    f"'tie_word_embeddings={cast(T5Config, self.config).tie_word_embeddings}'",
                    valid_options=[
                        "'is_gated_act' and 'tie_word_embeddings' should be different."
                    ],
                )

            if cast(T5Config, self.config).layer_norm_epsilon != 1e-6:
                raise NotSupportedCheckpointError(
                    invalid_option="'layer_norm_epsilon="
                    f"{cast(T5Config, self.config).layer_norm_epsilon}'",
                    valid_options=[1e-6],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def _decoder_final_ln_weight_reshape(
        self, params: List[torch.Tensor]
    ) -> torch.Tensor:
        """Special handle for T5."""
        assert len(params) == 1
        param = params[0]

        if cast(T5Config, self.config).tie_word_embeddings:
            param = param * (cast(T5Config, self.config).d_model ** -0.5)

        return param

    def pos_embed_weight_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """Reshape positional embedding weights in T5."""
        assert len(params) == 1
        return params[0]

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(T5Config, self.config)

        logger.warn(
            "The 'max_input_length' and 'max_output_length' fields are left blank as "
            "they cannot be automatically configured. "
            "Determine the 'max_input_length' and 'max_output_length' according to your "
            "needs. The T5 model does not rely on absolute position embeddings, "
            "allowing you to choose any suitable value."
        )

        eos_token_id = self.get_eos_token_id()
        decoder_start_token_id = self.get_decoder_start_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.encoder_head_size,
            "num_heads": self.encoder_num_attention_heads,
            "hidden_size": self.encoder_hidden_size,
            "ff_intermediate_size": self.decoder_ff_intermediate_size,
            "num_encoder_layers": self.encoder_layer_num,
            "num_decoder_layers": self.decoder_layer_num,
            "max_input_length": "FILL ME",
            "max_output_length": "FILL ME",
            "num_pos_emb_buckets": config.relative_attention_num_buckets,
            "max_pos_distance": config.relative_attention_max_distance,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
            "decoder_start_token": (
                decoder_start_token_id
                if decoder_start_token_id is not None
                else "FILL ME"
            ),
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        if cast(T5Config, self.config).is_gated_act:
            return "t5-v1_1"
        return "t5"

    @property
    def encoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in T5's encoder."""
        convert_info_list = []
        for i in range(self.encoder_layer_num):
            layer_prefix = f"{self.encoder_layer_prefix}{i}."
            converted_prefixe = f"{ENCODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.0.layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.1.layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}layer.0.SelfAttention.q.weight",
                            f"{layer_prefix}layer.0.SelfAttention.k.weight",
                            f"{layer_prefix}layer.0.SelfAttention.v.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.0.SelfAttention.o.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                ]
            )

            if cast(T5Config, self.config).is_gated_act:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.1.DenseReluDense.wi_0.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_gate/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.1.DenseReluDense.wi_1.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_fc/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.1.DenseReluDense.wo.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_proj/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                    ]
                )
            else:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.1.DenseReluDense.wi.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_fc/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.1.DenseReluDense.wo.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_proj/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                    ]
                )

        return convert_info_list

    @property
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in T5's decoder."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefixe = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.0.layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.1.layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.2.layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}ln_3/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}layer.0.SelfAttention.q.weight",
                            f"{layer_prefix}layer.0.SelfAttention.k.weight",
                            f"{layer_prefix}layer.0.SelfAttention.v.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.0.SelfAttention.o.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}layer.1.EncDecAttention.q.weight",
                            f"{layer_prefix}layer.1.EncDecAttention.k.weight",
                            f"{layer_prefix}layer.1.EncDecAttention.v.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}cross_attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}layer.1.EncDecAttention.o.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefixe}cross_attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                ]
            )

            if cast(T5Config, self.config).is_gated_act:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.2.DenseReluDense.wi_0.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_gate/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.2.DenseReluDense.wi_1.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_fc/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.2.DenseReluDense.wo.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_proj/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                    ]
                )
            else:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.2.DenseReluDense.wi.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_fc/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}layer.2.DenseReluDense.wo.weight"
                            ],
                            data_type=self.data_type,
                            converted_name=f"{converted_prefixe}mlp/c_proj/weight:0",
                            reshape_fn=self.linear_weight_reshape,
                        ),
                    ]
                )

        return convert_info_list

    @property
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The convert_info_list for non-transformer blocks in T5."""
        convert_info_list = [
            ConvertInfo(
                param_names=[f"shared.weight"],
                data_type=self.data_type,
                converted_name="wte/weight:0",
                reshape_fn=self.token_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=[
                    "encoder.block.0.layer.0.SelfAttention.relative_attention_bias.weight"
                ],
                data_type=ModelDataType.FP32,
                converted_name=f"{ENCODER_PREFIX}/wpe/weight:0",
                reshape_fn=self.pos_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=[
                    "decoder.block.0.layer.0.SelfAttention.relative_attention_bias.weight"
                ],
                data_type=ModelDataType.FP32,
                converted_name=f"{DECODER_PREFIX}/wpe/weight:0",
                reshape_fn=self.pos_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["encoder.final_layer_norm.weight"],
                data_type=self.data_type,
                converted_name=f"{ENCODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
            ConvertInfo(
                param_names=["decoder.final_layer_norm.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self._decoder_final_ln_weight_reshape,
            ),
        ]

        if not cast(T5Config, self.config).tie_word_embeddings:
            convert_info_list.append(
                ConvertInfo(
                    param_names=["lm_head.weight"],
                    data_type=self.data_type,
                    converted_name="head_fc/weight:0",
                    reshape_fn=self.head_weight_reshape,
                )
            )

        return convert_info_list

    @property
    def encoder_layer_prefix(self) -> str:
        """The layer name prefix used before T5 encoder's transformer block number."""
        return "encoder.block."

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before T5 decoder's transformer block number."""
        return "decoder.block."

    @property
    def encoder_layer_num(self) -> int:
        """The number of transformer blocks in T5 encoder."""
        return cast(T5Config, self.config).num_layers

    @property
    def encoder_hidden_size(self) -> int:
        """The hidden size of T5 encoder."""
        return cast(T5Config, self.config).d_model

    @property
    def encoder_num_attention_heads(self) -> int:
        """The number of attention heads of T5 encoder."""
        return cast(T5Config, self.config).num_heads

    @property
    def encoder_head_size(self) -> int:
        """The head size of T5 encoder."""
        return cast(T5Config, self.config).d_kv

    @property
    def encoder_ff_intermediate_size(self) -> int:
        """The intermediate of the linear layer in T5 encoder's MLP."""
        return cast(T5Config, self.config).d_ff

    @property
    def decoder_layer_num(self) -> int:
        """The number of transformer blocks in T5 decoder."""
        return cast(T5Config, self.config).num_decoder_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size of T5 decoder."""
        return cast(T5Config, self.config).d_model

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads of T5 decoder."""
        return cast(T5Config, self.config).num_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads of t5 decoder."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of T5 decoder."""
        return cast(T5Config, self.config).d_kv

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate of the linear layer in T5 decoder's MLP."""
        return cast(T5Config, self.config).d_ff



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli FalconForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Tuple, Type, cast

import torch
from transformers.models.falcon import (  # type: ignore[import]
    FalconConfig,
    FalconForCausalLM,
)

from friendli.modules.converter.base import OneOfConverter
from friendli.modules.converter.utils import convert_to_gpt_j_params
from friendli.modules.quantizer.schema.config import SmoothQuantConfig
from friendli.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from friendli.modules.quantizer.smoothquant.base import SmoothQuantHook


class SmoothQuantFalconHook(SmoothQuantHook):
    """SmoothQuant Hook for FalconForCausalLM."""

    def __init__(self, quant_config: Dict[str, Any], converter: OneOfConverter):
        """Initialize SmoothQuantFalconHook."""
        super().__init__(quant_config, converter)
        config = cast(FalconConfig, converter.config)
        self.num_attention_heads = config.num_attention_heads
        self.hidden_size = config.hidden_size
        self.head_size = self.hidden_size // self.num_attention_heads
        self.rotary_dim = self.head_size
        self.num_kv_attention_heads = self.get_num_kv_attention_heads(config)

    def get_num_kv_attention_heads(self, config: FalconConfig) -> int:
        """Returns the number of key-value attention heads in FalconForCausalLM."""
        if config.new_decoder_architecture:
            if config.num_kv_heads is not None:
                return config.num_kv_heads
            return config.num_attention_heads

        if config.multi_query:
            return 1

        if config.num_kv_heads is not None:
            return config.num_kv_heads
        return config.num_attention_heads

    def iter_smooth_norm_weights(
        self,
        model: FalconForCausalLM,
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm's weight and linear layer's weight per transformer block in FalconForCausalLM."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        for index, decoder_layer in enumerate(model.transformer.h):  # type: ignore[union-attr]
            if cast(FalconConfig, self.converter.config).new_decoder_architecture:
                # [LayerNorm 1] - [ QKV projection ] gets smoothed
                yield (
                    [
                        decoder_layer.ln_attn.weight.data,
                        decoder_layer.ln_attn.bias.data,
                    ],
                    [
                        decoder_layer.self_attention.query_key_value.weight.data,
                    ],
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",  # the input tensors fed into Q, K, V matrices are identical.
                )
                # [LayerNorm 2] - [ MLP FF 1 ] gets smoothed
                yield (
                    [
                        decoder_layer.ln_mlp.weight.data,
                        decoder_layer.ln_mlp.bias.data,
                    ],
                    [decoder_layer.mlp.dense_h_to_4h.weight.data],  # [OutDim, InDim]
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_h_to_4h",
                )
            else:
                # [LayerNorm 1] - [ QKV projection ] gets smoothed ( MLP FF1 is not smoothed. No LayerNorm 2. )
                yield (
                    [
                        decoder_layer.input_layernorm.weight.data,
                    ],
                    [
                        decoder_layer.self_attention.query_key_value.weight.data,
                    ],
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",  # the input tensors fed into Q, K, V matrices are identical.
                )

            if quant_args.attn_fc_smoothing:
                yield (
                    [decoder_layer.attn_fc_pre_smoother.scale.data],
                    [decoder_layer.self_attention.dense.weight.data],
                    f"{self.quantized_layer_prefix}{index}.self_attention.dense",
                )
            if quant_args.ff2_smoothing:
                yield (
                    [decoder_layer.ff2_pre_smoother.scale.data],
                    [decoder_layer.mlp.dense_4h_to_h.weight.data],
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_4h_to_h",
                )

    def reshape_qkv_weight(
        self, attn_layer: torch.nn.Module
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Reshapes the qkv weight in FalconForCausalLM for Quantization."""
        qkv_weight = cast(torch.nn.Linear, attn_layer.query_key_value).weight
        num_queries_per_kv = self.num_attention_heads // self.num_kv_attention_heads

        qkv_weight = qkv_weight.reshape(
            self.num_kv_attention_heads,
            num_queries_per_kv + 2,
            self.head_size,
            self.hidden_size,
        )

        q_weight = qkv_weight[:, :num_queries_per_kv].reshape(
            self.num_kv_attention_heads * num_queries_per_kv,
            self.head_size,
            self.hidden_size,
        )
        k_weight = qkv_weight[:, [-2]].reshape(
            self.num_kv_attention_heads,
            self.head_size,
            self.hidden_size,
        )
        v_weight = qkv_weight[:, [-1]].reshape(
            self.num_kv_attention_heads * self.head_size,
            self.hidden_size,
        )

        q_weight = convert_to_gpt_j_params(q_weight, self.rotary_dim)
        k_weight = convert_to_gpt_j_params(k_weight, self.rotary_dim)

        q_weight = q_weight.reshape(
            self.num_kv_attention_heads * num_queries_per_kv * self.head_size,
            self.hidden_size,
        )
        k_weight = k_weight.reshape(
            self.num_kv_attention_heads * self.head_size,
            self.hidden_size,
        )

        return q_weight, k_weight, v_weight

    def sort_qkv_output_stats(self, max_output_stat: torch.Tensor) -> torch.Tensor:
        """Sort max output stats of qkv_layer in FalconForCausalLM."""
        num_queries_per_kv = self.num_attention_heads // self.num_kv_attention_heads
        qkv_output_stat = max_output_stat.reshape(
            self.num_kv_attention_heads,
            num_queries_per_kv + 2,
            self.head_size,
        )
        q_out_stats = qkv_output_stat[:, :num_queries_per_kv].reshape(
            self.num_kv_attention_heads * num_queries_per_kv,
            self.head_size,
        )
        k_out_stats = qkv_output_stat[:, [-2]].reshape(
            self.num_kv_attention_heads,
            self.head_size,
        )
        v_out_stats = qkv_output_stat[:, [-1]].reshape(
            self.num_kv_attention_heads * self.head_size,
        )
        q_out_stats = convert_to_gpt_j_params(q_out_stats, self.rotary_dim)
        k_out_stats = convert_to_gpt_j_params(k_out_stats, self.rotary_dim)
        q_out_stats = q_out_stats.reshape(
            self.num_kv_attention_heads * num_queries_per_kv * self.head_size,
        )
        k_out_stats = k_out_stats.reshape(
            self.num_kv_attention_heads * self.head_size,
        )

        return torch.cat((q_out_stats, k_out_stats, v_out_stats), dim=0)

    def iter_tf_quant_inputs(self, model: FalconForCausalLM) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of FalconForCausalLM."""
        for index, decoder_layer in enumerate(model.transformer.h):
            self_attn = decoder_layer.self_attention
            q_weight, k_weight, v_weight = self.reshape_qkv_weight(self_attn)
            qkv_weight = torch.cat((q_weight, k_weight, v_weight), dim=0)
            fc1 = decoder_layer.mlp.dense_h_to_4h
            fc2 = decoder_layer.mlp.dense_4h_to_h

            yield TFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",
                    0,
                    q_weight.size(0),
                    self.sort_qkv_output_stats,
                ),
                k=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",
                    q_weight.size(0),
                    q_weight.size(0) + k_weight.size(0),
                    self.sort_qkv_output_stats,
                ),
                v=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",
                    q_weight.size(0) + k_weight.size(0),
                    qkv_weight.size(0),
                    self.sort_qkv_output_stats,
                ),
                attn_fc=QuantInput(
                    self_attn.dense.weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.dense",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_h_to_4h",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_4h_to_h",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in FalconForCausalLM."""
        return (torch.nn.Linear,)

    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after attention in the decoder layer."""
        return decoder_layer.self_attention.dense

    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after FF1 in the decoder layer."""
        return decoder_layer.mlp.dense_4h_to_h

    def get_tf_blocks(self, model: FalconForCausalLM) -> List[torch.nn.Module]:
        """Returns the decoder layers(transformer blocks) in the model."""
        return list(model.transformer.h)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli OPT Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Dict, List, cast

import torch
from transformers import OPTConfig  # type: ignore[import]

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import DECODER_PREFIX, DecoderOnlyConverter
from friendli.modules.converter.schema import ConvertInfo


class OPTForCausalLMConverter(DecoderOnlyConverter):
    """OPTForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if OPT architectures' config can be converted to Friendli format."""
        super().check_config()
        try:
            if cast(OPTConfig, self.config).activation_function not in ["relu"]:
                raise NotSupportedCheckpointError(
                    invalid_option="'activation_function="
                    f"{cast(OPTConfig, self.config).activation_function}'",
                    valid_options=["relu"],
                )
            if not cast(OPTConfig, self.config).do_layer_norm_before is True:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'do_layer_norm_before={False}'",
                    valid_options=[True],
                )
            if (
                cast(OPTConfig, self.config).word_embed_proj_dim
                != cast(OPTConfig, self.config).hidden_size
            ):
                raise NotSupportedCheckpointError(
                    invalid_option="'word_embed_proj_dim"
                    f"({cast(OPTConfig, self.config).word_embed_proj_dim}) "
                    f"!= hidden_size({cast(OPTConfig, self.config).hidden_size})'",
                    valid_options=[
                        f"'word_embed_proj_dim({cast(OPTConfig, self.config).hidden_size}) "
                        f"== hidden_size({cast(OPTConfig, self.config).hidden_size})'"
                    ],
                )
            if cast(  # pylint: disable=protected-access
                OPTConfig, self.config
            )._remove_final_layer_norm:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'_remove_final_layer_norm={True}'",
                    valid_options=[False],
                )
            if not cast(OPTConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'tie_word_embeddings={False}'",
                    valid_options=[True],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def pos_embed_weight_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """Positional embedding weight convert for OPT's decoder."""
        assert len(params) == 1
        pos_emb = params[0]
        pos_emb = pos_emb[2:, :]  # offset pos emb

        return pos_emb

    def qkv_weight_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """qkv_weight_reshape for OPT's attention layer."""
        qkv_weight = torch.cat(
            params,
            dim=0,
        )
        qkv_weight = qkv_weight.transpose(0, 1)
        return qkv_weight

    def qkv_bias_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """qkv_bias_reshape for OPT's attention layer."""
        qkv_bias = torch.cat(
            params,
            dim=0,
        )
        return qkv_bias

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(OPTConfig, self.config)

        logger.warn(
            "Since OPT uses absolute position embedding, 'max_length' cannot be "
            "larger than %d.",
            config.max_position_embeddings,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "num_heads": self.decoder_num_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": config.max_position_embeddings,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "opt"

    @property
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in OPT."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn_layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn_layer_norm.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.bias",
                            f"{layer_prefix}self_attn.k_proj.bias",
                            f"{layer_prefix}self_attn.v_proj.bias",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/bias:0",
                        reshape_fn=self.qkv_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.out_proj.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}final_layer_norm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}final_layer_norm.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc1.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc2.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc2.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}fc1.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.out_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.weight",
                            f"{layer_prefix}self_attn.k_proj.weight",
                            f"{layer_prefix}self_attn.v_proj.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                ]
            )
        return convert_info_list

    @property
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for non-transformer blocks in OPT."""
        return [
            ConvertInfo(
                param_names=["model.decoder.embed_tokens.weight"],
                data_type=self.data_type,
                converted_name="wte/weight:0",
                reshape_fn=self.token_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.decoder.embed_positions.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/wpe/weight:0",
                reshape_fn=self.pos_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.decoder.final_layer_norm.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.decoder.final_layer_norm.bias"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/beta:0",
                reshape_fn=self.ln_bias_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.weight"],
                data_type=self.data_type,
                converted_name="head_fc/weight:0",
                reshape_fn=self.head_weight_reshape,
            ),
        ]

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before OPT's transformer block number."""
        return "model.decoder.layers."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in OPT."""
        return cast(OPTConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in OPT."""
        return cast(OPTConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in OPT."""
        return cast(OPTConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in opt."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of OPT."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of the linear layer in codegen OPT."""
        return self.decoder_hidden_size * 4



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Checkpoint Converter."""

from __future__ import annotations

from typing import Optional

import yaml
from peft import PeftModel  # type: ignore[import] # pylint: disable=import-error

from friendli.enums import CheckpointFileType, ModelDataType
from friendli.errors import TokenizerNotFoundError
from friendli.logging import logger
from friendli.modules.converter.saver import get_saver
from friendli.utils.validate import validate_convert_imports

validate_convert_imports()
# pylint: disable=import-outside-toplevel, wrong-import-position, wrong-import-order, ungrouped-imports
import torch  # type: ignore[import]
from accelerate import init_empty_weights  # type: ignore[import]

from friendli.modules.converter.maps import (
    get_adapter_converter_factory,
    get_hf_converter_factory,
)
from friendli.modules.converter.utils import (
    get_adapter_config,
    get_model_arch,
    get_model_generation_config,
    get_model_pretrained_config,
    save_tokenizer,
)
from friendli.modules.quantizer.maps import get_quantized_converter
from friendli.modules.quantizer.schema.config import OneOfQuantConfig

# pylint: enable=import-outside-toplevel, wrong-import-position, wrong-import-order, ungrouped-imports


def convert_checkpoint(  # pylint: disable=too-many-locals
    model_name_or_path: str,
    model_output_path: str,
    data_type: ModelDataType,
    output_ckpt_file_type: CheckpointFileType,
    *,
    tokenizer_output_dir: Optional[str] = None,
    attr_output_path: Optional[str] = None,
    cache_dir: Optional[str] = None,
    dry_run: bool = False,
    quantize: bool = False,
    quant_config: Optional[OneOfQuantConfig] = None,
) -> None:
    """Convert HuggingFace model checkpoint to Friendli format.

    Args:
        model_name_or_path (str): Hugging Face model name or local path to the checkpoint.
        model_output_path (str): Path to the converted checkpoint to save.
        data_type (CheckpointDataType): Converted checkpoint data type.
        tokenizer_output_dir (Optional[str], optional): Directory path to save 'tokenizer.json'
            for the converted checkpoint. Defaults to None.
        attr_output_path (Optional[str], optional): Path to create the attribute YAML file for
            the converted checkpoint. Defaults to None.
        cache_dir (Optional[str], optional): Path for downloading checkpoint. Defaults to None.
        dry_run (bool, optional): Check only if checkpoint is convertable. Defaults to False.
        quantize (bool, optional): Enable quantization. Defaults to False.
        quant_config (Optional[OneOfQuantConfig], optional): Quantization configuration.
            Defaults to None.

    Raises:
        InValidconfigError: Raised when data_type is not supported.
        NotFoundError: Raised when `model_name_or_path` or `tokenizer_output_dir` is not found.
        NotSupportedCheckpointError: Raised when model architecture is not supported to convert.

    """
    model_config = get_model_pretrained_config(
        model_name_or_path, model_output_path, cache_dir
    )
    generation_config = get_model_generation_config(model_name_or_path, cache_dir)

    model_arch = get_model_arch(model_config)
    hf_factory, converter_factory = get_hf_converter_factory(model_arch)
    converter = converter_factory(
        config=model_config,
        generation_config=generation_config,
        data_type=data_type,
    )

    if quantize:
        assert quant_config is not None
        converter = get_quantized_converter(  # type: ignore[assignment]
            quant_config, converter
        )

    converter.check_config()

    if not dry_run:
        logger.info(
            "Start loading Hugging Face checkpoint(%s) for conversion...",
            model_name_or_path,
        )
        model = hf_factory.from_pretrained(
            model_name_or_path,
            torch_dtype=torch.float32,
            cache_dir=cache_dir,
            trust_remote_code=True,
            low_cpu_mem_usage=True,
            # `low_cpu_mem_usage` is for model loading faster and using ~1x model size CPU memory.
            # https://huggingface.co/docs/transformers/main_classes/model#transformers.PreTrainedModel.from_pretrained.example
        )

        logger.info(
            "Hugging Face checkpoint(%s) is successfully loaded!",
            model_name_or_path,
        )

        convert_info_list = converter.get_convert_info_list()
        with get_saver(output_ckpt_file_type, model_output_path) as saver:
            for name, w in converter.convert(model, convert_info_list):
                saver.save_tensor(name, w)

        logger.info(
            "Hugging Face checkpoint(%s) is successfully converted to Friendli format!",
            model_name_or_path,
        )

    if attr_output_path is not None:
        attr = converter.get_attributes()
        with open(attr_output_path, "w", encoding="utf-8") as file:
            yaml.dump(attr, file, sort_keys=False)

    if tokenizer_output_dir is not None:
        try:
            save_tokenizer(
                model_name_or_path=model_name_or_path,
                cache_dir=cache_dir,
                save_dir=tokenizer_output_dir,
            )
        except TokenizerNotFoundError as exc:
            logger.warn(str(exc))


def convert_adapter_checkpoint(  # pylint: disable=too-many-locals
    adapter_name_or_path: str,
    adapter_output_path: str,
    adapter_attr_output_path: str,
    data_type: ModelDataType,
    output_adapter_file_type: CheckpointFileType,
    cache_dir: Optional[str],
    dry_run: bool = False,
) -> None:
    """Convert HuggingFace model checkpoint to Friendli format."""
    adapter_config = get_adapter_config(adapter_name_or_path, cache_dir)
    model_config = get_model_pretrained_config(
        adapter_config.base_model_name_or_path, adapter_attr_output_path, cache_dir
    )
    model_arch = get_model_arch(model_config)
    hf_factory, converter_factory = get_hf_converter_factory(model_arch)
    converter = converter_factory(
        config=model_config,
        generation_config=None,
        data_type=data_type,
    )
    adapter_converter = get_adapter_converter_factory(model_arch)(
        converter, adapter_config
    )
    adapter_converter.check_config()

    if not dry_run:
        logger.info(
            "Start loading Hugging Face adapter checkpoint(%s's %s) for conversion...",
            adapter_config.base_model_name_or_path,
            adapter_name_or_path,
        )
        with init_empty_weights():
            model = hf_factory.from_pretrained(
                adapter_config.base_model_name_or_path,
                torch_dtype=torch.float32,
                cache_dir=cache_dir,
                trust_remote_code=True,
                low_cpu_mem_usage=True,
            )
        # inplace model update
        PeftModel.from_pretrained(
            model, adapter_name_or_path, cache_dir=cache_dir, torch_dtype=torch.float32
        )
        logger.info(
            "Hugging Face adapter checkpoint (%s) is successfully loaded!",
            adapter_name_or_path,
        )
        convert_dict = adapter_converter.get_convert_info_list()
        with get_saver(output_adapter_file_type, adapter_output_path) as saver:
            for name, w in adapter_converter.convert(model, convert_dict):
                saver.save_tensor(name, w)

        if adapter_attr_output_path is not None:
            attr = adapter_converter.get_attributes()
            with open(adapter_attr_output_path, "w", encoding="utf-8") as file:
                yaml.dump([attr], file, sort_keys=False)

        logger.info(
            "Hugging Face checkpoint (%s) is successfully converted to Friendli format!",
            adapter_name_or_path,
        )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Quantizer Maps."""

from __future__ import annotations

from typing import Any, Dict, Type

from friendli.enums import QuantMode
from friendli.errors import NotSupportedQuantModeError
from friendli.modules.converter.base import OneOfConverter
from friendli.modules.converter.utils import get_model_arch
from friendli.modules.quantizer.awq.base import AWQHook, AWQQuantizer
from friendli.modules.quantizer.awq.models.gpt_neox import AWQGPTNeoXHook
from friendli.modules.quantizer.awq.models.gptj import AWQGPTJHook
from friendli.modules.quantizer.awq.models.llama import AWQLlamaHook
from friendli.modules.quantizer.awq.models.mpt import AWQMPTHook
from friendli.modules.quantizer.base import CommonQuantizer
from friendli.modules.quantizer.schema.config import OneOfQuantConfig
from friendli.modules.quantizer.smoothquant.base import (
    SmoothQuantHook,
    SmoothQuantQuantizer,
)
from friendli.modules.quantizer.smoothquant.models.bloom import SmoothQuantBloomHook
from friendli.modules.quantizer.smoothquant.models.codegen import SmoothQuantCodeGenHook
from friendli.modules.quantizer.smoothquant.models.falcon import SmoothQuantFalconHook
from friendli.modules.quantizer.smoothquant.models.gpt2 import SmoothQuantGPT2Hook
from friendli.modules.quantizer.smoothquant.models.gpt_neox import (
    SmoothQuantGPTNeoXHook,
)
from friendli.modules.quantizer.smoothquant.models.gptj import SmoothQuantGPTJHook
from friendli.modules.quantizer.smoothquant.models.llama import SmoothQuantLlamaHook
from friendli.modules.quantizer.smoothquant.models.mpt import SmoothQuantMPTHook
from friendli.modules.quantizer.smoothquant.models.opt import SmoothQuantOPTHook

model_arch_smoothquant_hook_map: Dict[str, type[SmoothQuantHook]] = {
    "OPTForCausalLM": SmoothQuantOPTHook,
    "MPTForCausalLM": SmoothQuantMPTHook,
    "BloomForCausalLM": SmoothQuantBloomHook,
    "CodeGenForCausalLM": SmoothQuantCodeGenHook,
    "GPTNeoXForCausalLM": SmoothQuantGPTNeoXHook,
    "GPTJForCausalLM": SmoothQuantGPTJHook,
    "GPT2LMHeadModel": SmoothQuantGPT2Hook,
    "FalconForCausalLM": SmoothQuantFalconHook,
    "LlamaForCausalLM": SmoothQuantLlamaHook,
}

model_arch_awq_hook_map: Dict[str, type[AWQHook]] = {
    "GPTJForCausalLM": AWQGPTJHook,
    "GPTNeoXForCausalLM": AWQGPTNeoXHook,
    "LlamaForCausalLM": AWQLlamaHook,
    "MPTForCausalLM": AWQMPTHook,
    "MistralForCausalLM": AWQLlamaHook,
}


def get_quanthook_map(quant_mode: QuantMode) -> Dict[str, Any]:
    """Get quantizer map."""
    if quant_mode == QuantMode.SMOOTH_QUANT:
        return model_arch_smoothquant_hook_map
    if quant_mode == QuantMode.AWQ:
        return model_arch_awq_hook_map
    raise NotSupportedQuantModeError(
        invalid_option=quant_mode,
        valid_options=[e.value for e in QuantMode],
    )


def get_quantizer_class(quant_mode: QuantMode) -> Type[CommonQuantizer]:
    """Get quantizer class."""
    if quant_mode == QuantMode.SMOOTH_QUANT:
        return SmoothQuantQuantizer
    if quant_mode == QuantMode.AWQ:
        return AWQQuantizer
    raise NotSupportedQuantModeError(
        invalid_option=quant_mode,
        valid_options=[e.value for e in QuantMode],
    )


def get_quantized_converter(
    quant_config: OneOfQuantConfig,
    converter: OneOfConverter,
) -> CommonQuantizer:
    """Get quantizer for specific model architecture with quant mode and args."""
    model_arch = get_model_arch(converter.config)
    quant_mode = quant_config.mode
    quantizer = get_quantizer_class(quant_mode)
    quanthook_map = get_quanthook_map(quant_mode)
    quanthook = quanthook_map[model_arch](quant_config, converter)
    return quantizer(quanthook, quant_config, converter)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Object Transfer Utils."""

# pylint: disable=too-many-locals, too-many-arguments


from __future__ import annotations

import heapq
import math
import os
import socket
from concurrent.futures import FIRST_EXCEPTION, Executor, ThreadPoolExecutor, wait
from functools import wraps
from pathlib import Path
from threading import Lock
from typing import Any, Callable, Dict, List, Optional, Set

import requests
from requests import Request, Session
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper
from urllib3.exceptions import ReadTimeoutError

from friendli.errors import (
    InvalidPathError,
    MaxRetriesExceededError,
    NotFoundError,
    TransferError,
)
from friendli.logging import logger
from friendli.schema.resource.v1.transfer import (
    MultipartUploadTask,
    MultipartUploadUrlInfo,
    UploadedPartETag,
    UploadTask,
)
from friendli.utils.compat import model_dump
from friendli.utils.fs import get_file_size, storage_path_to_local_path
from friendli.utils.request import DEFAULT_REQ_TIMEOUT

KiB = 1024
MiB = KiB * KiB
GiB = MiB * KiB
IO_CHUNK_SIZE = 256 * KiB
S3_MULTIPART_THRESHOLD = 16 * MiB
S3_MULTIPART_CHUNK_SIZE = 16 * MiB
S3_MAX_SINGLE_UPLOAD_SIZE = 5 * GiB
NUM_MAX_PARTS = 10000
S3_RETRYABLE_DOWNLOAD_ERRORS = (
    socket.timeout,
    ConnectionError,
    requests.exceptions.ReadTimeout,
    requests.exceptions.ConnectionError,
    ReadTimeoutError,
)
MAX_RETRIES = 5
MAX_DEFAULT_REQUEST_CONFIG_SIZE = GiB


class DownloadManager:
    """Download manager."""

    def __init__(
        self,
        write_queue: DeferQueue,
        io_chunk_size: int = IO_CHUNK_SIZE,
        multipart_threshold: int = S3_MULTIPART_THRESHOLD,
        max_part_size: int = S3_MULTIPART_CHUNK_SIZE,
        max_retries: int = MAX_RETRIES,
    ) -> None:
        """Initializes DownloadManager."""
        self._write_queue = write_queue
        self._io_chunk_size = io_chunk_size
        self._multipart_threshold = multipart_threshold
        self._max_part_size = max_part_size
        self._max_retries = max_retries
        self._io_lock = Lock()

    def download_file(self, url: str, out: Path) -> None:
        """Download a file from the URL."""
        file_size = self._get_content_size(url)

        # Create directory if not exists
        dirpath = os.path.dirname(out)
        try:
            os.makedirs(dirpath, exist_ok=True)
        except OSError as exc:
            raise InvalidPathError(
                f"Cannot create directory({dirpath}) to download file: {exc!r}"
            ) from exc

        if file_size < self._multipart_threshold:
            self._download_file_sequential(url, out, file_size)
        else:
            self._download_file_parallel(url, out, file_size)

    def _download_file_sequential(
        self, url: str, out: Path, content_length: int
    ) -> None:
        """Downloads a file sequentially."""
        response = requests.get(url, stream=True, timeout=DEFAULT_REQ_TIMEOUT)

        with tqdm.wrapattr(
            open(out, "wb"),
            "write",
            desc=out.name,
            miniters=1,
            total=content_length,
        ) as fout:
            for chunk in response.iter_content(self._io_chunk_size):
                fout.write(chunk)

    def _download_file_parallel(self, url: str, out: Path, content_length: int) -> None:
        """Downloads a file in parallel."""
        chunks = range(0, content_length, self._max_part_size)

        with tqdm(
            desc=out.name,
            total=content_length,
            unit="B",
            unit_scale=True,
            unit_divisor=KiB,
        ) as pbar:
            with ThreadPoolExecutor() as executor:
                futs = {
                    executor.submit(
                        self._download_range,
                        url,
                        start,
                        start + self._max_part_size - 1,
                        out,
                        pbar,
                    )
                    for i, start in enumerate(chunks)
                }
                not_done = futs
                try:
                    while not_done:
                        done, not_done = wait(
                            not_done, timeout=1, return_when=FIRST_EXCEPTION
                        )
                        for fut in done:
                            fut.result()
                except KeyboardInterrupt as exc:
                    logger.warn(
                        "Keyboard interrupted. Wait a few seconds for the shutdown."
                    )
                    # py38 does not support cancel_futures option.
                    # Add cancel_futures=True after deprecation.
                    for fut in not_done:
                        fut.cancel()
                    executor.shutdown(wait=False)
                    raise exc

    def _download_range(
        self, url: str, start: int, end: int, out: Path, pbar: tqdm
    ) -> None:
        """Download a specific part of a file from the URL."""
        headers = {"Range": f"bytes={start}-{end}"}
        final_exc = None
        for i in range(self._max_retries):
            try:
                response = requests.get(
                    url, headers=headers, stream=True, timeout=DEFAULT_REQ_TIMEOUT
                )
                final_exc = None
                break
            except S3_RETRYABLE_DOWNLOAD_ERRORS as exc:
                logger.debug(
                    (
                        "Connection error while downloading. "
                        "Retry downloading the part (attempt %s / %s)."
                    ),
                    i + 1,
                    self._max_retries,
                )
                final_exc = exc
                continue

        if final_exc is not None:
            raise MaxRetriesExceededError(final_exc)

        downloaded_iter = response.iter_content(self._io_chunk_size)
        inner_offset = 0
        while True:
            final_exc = None
            for i in range(self._max_retries):
                try:
                    part = next(downloaded_iter)
                    pbar.update(len(part))
                    with self._io_lock:
                        writes = self._write_queue.request_writes(
                            offset=start + inner_offset, data=part
                        )
                        for write in writes:
                            with open(out, "ab") as f:
                                f.write(write)

                    inner_offset += len(part)
                    final_exc = None
                    break
                except StopIteration:
                    return
                except S3_RETRYABLE_DOWNLOAD_ERRORS as exc:
                    logger.debug(
                        (
                            "Connection error while downloading. "
                            "Retry downloading the part (attempt %s / %s)."
                        ),
                        i + 1,
                        self._max_retries,
                    )
                    final_exc = exc
                    continue

            if final_exc is not None:
                raise MaxRetriesExceededError(final_exc)

    def _get_content_size(self, url: str) -> int:
        """Get download content size."""
        response = requests.get(url, stream=True, timeout=DEFAULT_REQ_TIMEOUT)
        if response.status_code != 200:
            raise NotFoundError("Invalid presigned url")
        return int(response.headers["Content-Length"])


class UploadManager:
    """Upload manager."""

    def __init__(self, executor: Executor, chunk_adjuster: ChunksizeAdjuster) -> None:
        """Initializes UploadManager."""
        self._executor = executor
        self._adjuster = chunk_adjuster

    @classmethod
    def list_multipart_upload_objects(cls, path: Path) -> List[Path]:
        """Lists file paths that requires multipart upload under the given path."""
        if path.is_file():
            paths = [path] if get_file_size(path) >= S3_MULTIPART_THRESHOLD else []
        else:
            paths = [
                p
                for p in path.rglob("*")
                if p.is_file() and get_file_size(p) >= S3_MULTIPART_THRESHOLD
            ]
        return paths

    @classmethod
    def list_upload_objects(cls, path: Path) -> List[Path]:
        """Lists file paths that requires upload under the given path."""
        if path.is_file():
            paths = [path] if 0 < get_file_size(path) < S3_MULTIPART_THRESHOLD else []
        else:
            paths = [
                p
                for p in path.rglob("*")
                if p.is_file() and 0 < get_file_size(p) < S3_MULTIPART_THRESHOLD
            ]
        return paths

    def upload_file(
        self,
        upload_task: UploadTask,
        source_path: Optional[Path] = None,
    ) -> None:
        """Uploads a file in the local file system to Friendli."""
        if source_path is not None:
            local_path = storage_path_to_local_path(upload_task.path, source_path)
        else:
            local_path = Path(upload_task.path)

        file_size = get_file_size(local_path)

        with tqdm(
            desc=local_path.name,
            total=file_size,
            unit="B",
            unit_scale=True,
            unit_divisor=KiB,
        ) as pbar:
            try:
                with open(local_path, "rb") as f:
                    wrapped_object = CallbackIOWrapper(pbar.update, f, "read")
                    with Session() as s:
                        req = Request(
                            "PUT", upload_task.upload_url, data=wrapped_object
                        )
                        prep = req.prepare()
                        prep.headers["Content-Length"] = str(
                            file_size
                        )  # necessary to use ``CallbackIOWrapper``
                        response = s.send(prep)
                    if response.status_code != 200:
                        raise TransferError(
                            f"Failed to upload file ({local_path}): {response.content!r}"
                        )
            except FileNotFoundError as exc:
                raise NotFoundError(f"File '{local_path}' is not found.") from exc

    def multipart_upload_file(
        self,
        upload_task: MultipartUploadTask,
        source_path: Path,
        complete_callback: Callable[[str, str, List[Dict[str, Any]]], None],
        abort_callback: Callable[[str, str], None],
    ) -> None:
        """Uploads a file in the local file system to Friendli in multi-part."""
        local_path = storage_path_to_local_path(upload_task.path, source_path)
        file_size = get_file_size(local_path)
        total_num_parts = len(upload_task.upload_urls)
        uploaded_part_etags = []
        chunk_size = self._adjuster.adjust_chunksize(
            current_chunksize=S3_MULTIPART_CHUNK_SIZE, file_size=file_size
        )

        with tqdm(
            desc=local_path.name,
            total=file_size,
            unit="B",
            unit_scale=True,
            unit_divisor=KiB,
        ) as pbar:
            futs = {
                self._executor.submit(
                    self._upload_part,
                    file_path=local_path,
                    chunk_size=chunk_size,
                    url_info=url_info,
                    pbar=pbar,
                    is_last_part=(idx == total_num_parts - 1),
                )
                for idx, url_info in enumerate(upload_task.upload_urls)
            }
            not_done = futs
            try:
                while not_done:
                    done, not_done = wait(
                        not_done, timeout=1, return_when=FIRST_EXCEPTION
                    )
                    for fut in done:
                        part_etag = fut.result()
                        uploaded_part_etags.append(model_dump(part_etag))
                complete_callback(
                    upload_task.path, upload_task.upload_id, uploaded_part_etags
                )
            except KeyboardInterrupt:
                logger.warn(
                    "Keyboard interrupted. Wait a few seconds for the shutdown."
                )
                # py38 does not support cancel_futures option.
                # Add cancel_futures=True after deprecation.
                for fut in not_done:
                    fut.cancel()
                self._executor.shutdown(wait=False)
                abort_callback(upload_task.path, upload_task.upload_id)
                raise
            except Exception as exc:
                abort_callback(upload_task.path, upload_task.upload_id)
                raise TransferError(str(exc)) from exc

    def _upload_part(
        self,
        file_path: Path,
        chunk_size: int,
        url_info: MultipartUploadUrlInfo,
        pbar: tqdm,
        is_last_part: bool = False,
    ) -> UploadedPartETag:
        with open(file_path, "rb") as f:
            cursor = (url_info.part_number - 1) * chunk_size
            f.seek(cursor)

            file_size = get_file_size(file_path)
            chunk_size = min(chunk_size, file_size - cursor)
            wrapped_object = CustomCallbackIOWrapper(pbar.update, f, "read", chunk_size)

            with Session() as s:
                req = Request("PUT", url_info.upload_url, data=wrapped_object)
                prep = req.prepare()
                prep.headers["Content-Length"] = str(chunk_size)
                response = s.send(prep)
            response.raise_for_status()

            if is_last_part:
                if f.read(chunk_size):
                    raise TransferError(
                        "Some parts of your data is not uploaded. Please try again."
                    )

        return UploadedPartETag(
            etag=response.headers["ETag"], part_number=url_info.part_number
        )


class CustomCallbackIOWrapper(CallbackIOWrapper):
    """Cutom callback IO wrapper."""

    def __init__(self, callback, stream, method="read", chunk_size=None) -> None:
        """Initialize custom callback IO wrapper."""
        # Wrap a file-like object's `read` or `write` to report data length to the `callback`.
        super().__init__(callback, stream, method)
        self._chunk_size = chunk_size
        self._cursor = 0

        func = getattr(stream, method)
        if method == "write":

            @wraps(func)
            def write(data, *args, **kwargs):
                res = func(data, *args, **kwargs)
                callback(len(data))
                return res

            self.wrapper_setattr("write", write)
        elif method == "read":

            @wraps(func)
            def read(*args, **kwargs):
                assert chunk_size is not None
                if self._cursor >= chunk_size:
                    self._cursor = 0
                    return None

                data = func(*args, **kwargs)
                data_size = len(data)  # default to 8 KiB
                callback(data_size)
                self._cursor += data_size
                return data

            self.wrapper_setattr("read", read)
        else:
            raise KeyError("Can only wrap read/write methods")


# We modifed the original implementation of DeferQueue at boto/s3transfer.
# See https://github.com/boto/s3transfer.
class DeferQueue:
    """IO queue that defers the file writes until they are queued sequentially."""

    def __init__(self) -> None:
        """Initializes DeferQueue."""
        self._writes: List[Any] = []
        self._pending_offsets: Set[int] = set()
        self._next_offset: int = 0

    def request_writes(self, offset: int, data: Any) -> List[Any]:
        """Requests any available writes given new incoming data.

        You call this method by providing new data along with the offset associated with
        the data. If that new data unlocks any contiguous writes that can now be
        submitted, this method will return all applicable writes.

        This is done with 1 method call so you don't have to make two method calls
        (put(), get()) which acquires a lock each method call.

        """
        if offset < self._next_offset:
            # This is a request for a write that we've already seen. This can happen in
            # the event of a retry.
            return []
        if offset in self._pending_offsets:
            # We've already queued this offset so this request is a duplicate. In this
            # case we should ignore this request and prefer what's already queued.
            return []

        heapq.heappush(self._writes, (offset, data))
        self._pending_offsets.add(offset)

        writes = []
        while self._writes and self._writes[0][0] == self._next_offset:
            commit_offset, commit_data = heapq.heappop(self._writes)
            writes.append(commit_data)
            self._pending_offsets.remove(commit_offset)
            self._next_offset += len(commit_data)
        return writes


# We modifed the original implementation of ChunksizeAdjuster at boto/s3transfer.
# See https://github.com/boto/s3transfer.
class ChunksizeAdjuster:
    """Upload chunk size adjuster."""

    def __init__(
        self,
        max_size: int = S3_MAX_SINGLE_UPLOAD_SIZE,
        min_size: int = S3_MULTIPART_THRESHOLD,
        max_parts: int = NUM_MAX_PARTS,
    ) -> None:
        """Initializes ChunksizeAdjuster."""
        self._max_size = max_size
        self._min_size = min_size
        self._max_parts = max_parts

    def adjust_chunksize(
        self, current_chunksize: int, file_size: Optional[int] = None
    ) -> int:
        """Get a chunksize close to current that fits within all S3 limits.

        Args:
            current_chunksize (int): The currently configured chunksize.
            file_size (Optional[int], optional): The size of the file to upload. This
                might be None if the object being transferred has an unknown size.
                Defaults to None.

        Returns:
            int: A valid chunksize that fits within configured limits.

        """
        chunksize = current_chunksize
        if file_size is not None:
            chunksize = self._adjust_for_max_parts(chunksize, file_size)
        return self._adjust_for_chunksize_limits(chunksize)

    def _adjust_for_chunksize_limits(self, current_chunksize: int) -> int:
        if current_chunksize > self._max_size:
            logger.debug(
                "Chunksize greater than maximum chunksize. Setting to %s from %s.",
                self._max_size,
                current_chunksize,
            )
            return self._max_size
        if current_chunksize < self._min_size:
            logger.debug(
                "Chunksize less than minimum chunksize. Setting to %s from %s.",
                self._min_size,
                current_chunksize,
            )
            return self._min_size
        return current_chunksize

    def _adjust_for_max_parts(self, current_chunksize: int, file_size: int) -> int:
        chunksize = current_chunksize
        num_parts = int(math.ceil(file_size / float(chunksize)))

        while num_parts > self._max_parts:
            chunksize *= 2
            num_parts = int(math.ceil(file_size / float(chunksize)))

        if chunksize != current_chunksize:
            logger.debug(
                (
                    "Chunksize would result in the number of parts exceeding the "
                    "maximum. Setting to %s from %s."
                ),
                chunksize,
                current_chunksize,
            )

        return chunksize
